# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

from power_grid_model import ComponentType, DatasetType, power_grid_meta_data


def assert_data_type(pgm_meta_data_types, data_type):
    pgm_types = [pgm_type for pgm_type in pgm_meta_data_types]
    pgm_types.sort()
    generated_types = [member.value for member in data_type]
    generated_types.sort()
    assert pgm_types == generated_types


def test_power_grid_data_types():
    assert_data_type(power_grid_meta_data, DatasetType)


def test_power_grid_components():
    assert_data_type(power_grid_meta_data[DatasetType.input], ComponentType)



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""
Power Grid Model Validation Functions.

Although all functions are 'public', you probably only need validate_input_data() and validate_batch_data().

"""

import copy
from collections.abc import Sized as ABCSized
from itertools import chain
from typing import Literal

import numpy as np

from power_grid_model._core.dataset_definitions import ComponentType, DatasetType, _map_to_component_types
from power_grid_model._core.power_grid_meta import power_grid_meta_data
from power_grid_model._core.utils import (
    compatibility_convert_row_columnar_dataset as _compatibility_convert_row_columnar_dataset,
    convert_batch_dataset_to_batch_list as _convert_batch_dataset_to_batch_list,
)
from power_grid_model.data_types import BatchDataset, Dataset, SingleDataset
from power_grid_model.enum import (
    AngleMeasurementType,
    Branch3Side,
    BranchSide,
    CalculationType,
    FaultPhase,
    FaultType,
    LoadGenType,
    MeasuredTerminalType,
    WindingType,
)
from power_grid_model.validation._rules import (
    all_between as _all_between,
    all_between_or_at as _all_between_or_at,
    all_boolean as _all_boolean,
    all_cross_unique as _all_cross_unique,
    all_enabled_identical as _all_enabled_identical,
    all_finite as _all_finite,
    all_greater_or_equal as _all_greater_or_equal,
    all_greater_than_or_equal_to_zero as _all_greater_than_or_equal_to_zero,
    all_greater_than_zero as _all_greater_than_zero,
    all_in_valid_values as _all_in_valid_values,
    all_less_than as _all_less_than,
    all_not_two_values_equal as _all_not_two_values_equal,
    all_not_two_values_zero as _all_not_two_values_zero,
    all_same_current_angle_measurement_type_on_terminal as _all_same_current_angle_measurement_type_on_terminal,
    all_same_sensor_type_on_same_terminal as _all_same_sensor_type_on_same_terminal,
    all_unique as _all_unique,
    all_valid_associated_enum_values as _all_valid_associated_enum_values,
    all_valid_clocks as _all_valid_clocks,
    all_valid_enum_values as _all_valid_enum_values,
    all_valid_fault_phases as _all_valid_fault_phases,
    all_valid_ids as _all_valid_ids,
    any_voltage_angle_measurement_if_global_current_measurement as _any_voltage_angle_measurement_if_global_current_measurement,  # noqa: E501
    ids_valid_in_update_data_set as _ids_valid_in_update_data_set,
    no_strict_subset_missing as _no_strict_subset_missing,
    none_missing as _none_missing,
    not_all_missing as _not_all_missing,
    valid_p_q_sigma as _valid_p_q_sigma,
)
from power_grid_model.validation.errors import (
    IdNotInDatasetError,
    InvalidIdError,
    MissingValueError,
    MultiComponentNotUniqueError,
    ValidationError,
)
from power_grid_model.validation.utils import _update_input_data


def validate_input_data(
    input_data: SingleDataset, calculation_type: CalculationType | None = None, symmetric: bool = True
) -> list[ValidationError] | None:
    """
    Validates the entire input dataset:

        1. Is the data structure correct? (checking data types and numpy array shapes)
        2. Are all required values provided? (checking NaNs)
        3. Are all ID's unique? (checking object identifiers across all components)
        4. Are the supplied values valid? (checking limits and other logic as described in "Graph Data Model")

    Args:
        input_data: A power-grid-model input dataset
        calculation_type: Supply a calculation method, to allow missing values for unused fields
        symmetric: A boolean to state whether input data will be used for a symmetric or asymmetric calculation

    Returns:
        None if the data is valid, or a list containing all validation errors.

    Raises:
        Error: KeyError | TypeError | ValueError: if the data structure is invalid.
    """
    input_data = _map_to_component_types(input_data)

    # Convert to row based if in columnar or mixed format format
    row_input_data = _compatibility_convert_row_columnar_dataset(input_data, None, DatasetType.input)

    # A deep copy is made of the input data, since default values will be added in the validation process
    input_data_copy = copy.deepcopy(row_input_data)

    assert_valid_data_structure(input_data_copy, DatasetType.input)

    errors: list[ValidationError] = []
    errors += validate_required_values(input_data_copy, calculation_type, symmetric)
    errors += validate_unique_ids_across_components(input_data_copy)
    errors += validate_values(input_data_copy, calculation_type)
    return errors if errors else None


def validate_batch_data(
    input_data: SingleDataset,
    update_data: BatchDataset,
    calculation_type: CalculationType | None = None,
    symmetric: bool = True,
) -> dict[int, list[ValidationError]] | None:
    """
    The input dataset is validated:

        1. Is the data structure correct? (checking data types and numpy array shapes)
        2. Are all input data ID's unique? (checking object identifiers across all components)

    For each batch the update data is validated:
        3. Is the update data structure correct? (checking data types and numpy array shapes)
        4. Are all update ID's valid? (checking object identifiers across update and input data)

    Then (for each batch independently) the input dataset is updated with the batch's update data and validated:
        5. Are all required values provided? (checking NaNs)
        6. Are the supplied values valid? (checking limits and other logic as described in "Graph Data Model")

    Args:
        input_data: A power-grid-model input dataset
        update_data: A power-grid-model update dataset (one or more batches)
        calculation_type: Supply a calculation method, to allow missing values for unused fields
        symmetric: A boolean to state whether input data will be used for a symmetric or asymmetric calculation

    Returns:
        None if the data is valid, or a dictionary containing all validation errors,
        where the key is the batch number (0-indexed).

    Raises:
        Error: KeyError | TypeError | ValueError: if the data structure is invalid.
    """
    input_data = _map_to_component_types(input_data)
    update_data = _map_to_component_types(update_data)

    # Convert to row based if in columnar or mixed format
    row_input_data = _compatibility_convert_row_columnar_dataset(input_data, None, DatasetType.input)

    # A deep copy is made of the input data, since default values will be added in the validation process
    input_data_copy = copy.deepcopy(row_input_data)
    assert_valid_data_structure(input_data_copy, DatasetType.input)

    input_errors: list[ValidationError] = list(validate_unique_ids_across_components(input_data_copy))

    batch_data = _convert_batch_dataset_to_batch_list(update_data, DatasetType.update)

    errors = {}
    for batch, batch_update_data in enumerate(batch_data):
        row_update_data = _compatibility_convert_row_columnar_dataset(batch_update_data, None, DatasetType.update)
        assert_valid_data_structure(row_update_data, DatasetType.update)
        id_errors: list[IdNotInDatasetError | InvalidIdError] = validate_ids(row_update_data, input_data_copy)

        batch_errors = input_errors + id_errors

        if not id_errors:
            merged_data = _update_input_data(input_data_copy, row_update_data)
            batch_errors += validate_required_values(merged_data, calculation_type, symmetric)
            batch_errors += validate_values(merged_data, calculation_type)

        if batch_errors:
            errors[batch] = batch_errors

    return errors if errors else None


def assert_valid_data_structure(data: Dataset, data_type: DatasetType) -> None:
    """
    Checks if all component names are valid and if the data inside the component matches the required Numpy
    structured array as defined in the Power Grid Model meta data.

    Args:
        data: A power-grid-model input/update dataset
        data_type: 'input' or 'update'

    Raises:
        Error: KeyError, TypeError

    """
    if data_type not in {DatasetType.input, DatasetType.update}:
        raise KeyError(f"Unexpected data type '{data_type}' (should be 'input' or 'update')")

    component_dtype = {component: meta.dtype for component, meta in power_grid_meta_data[data_type].items()}
    for component, array in data.items():
        # Check if component name is valid
        if component not in component_dtype:
            raise KeyError(f"Unknown component '{component}' in {data_type} data.")

        # Check if component definition is as expected
        dtype = component_dtype[component]
        if isinstance(array, np.ndarray):
            if array.dtype != dtype:
                if not hasattr(array.dtype, "names") or not array.dtype.names:
                    raise TypeError(
                        f"Unexpected Numpy array ({array.dtype}) for '{component}' {data_type} data "
                        "(should be a Numpy structured array)."
                    )
                raise TypeError(
                    f"Unexpected Numpy structured array; (expected = {dtype}, actual = {array.dtype}). "
                    f"For component '{component}'."
                )
        else:
            raise TypeError(
                f"Unexpected data type {type(array).__name__} for '{component}' {data_type} data "
                "(should be a Numpy structured array)."
            )


def validate_unique_ids_across_components(data: SingleDataset) -> list[MultiComponentNotUniqueError]:
    """
    Checks if all ids in the input dataset are unique

    Args:
        data: A power-grid-model input dataset

    Returns:
        An empty list if all ids are unique, or a list of MultiComponentNotUniqueErrors for all components that
        have non-unique ids
    """
    return _all_cross_unique(data, [(component, "id") for component in data])


def validate_ids(update_data: SingleDataset, input_data: SingleDataset) -> list[IdNotInDatasetError | InvalidIdError]:
    """
    Checks if all ids of the components in the update data:
     - exist and match those in the input data
     - are not present but qualifies for optional id

    This function should be called for every update dataset in a batch set

    Args:
        update_data: A single update dataset
        input_data: Input dataset

    Returns:
        An empty list if all update data ids are valid, or a list of IdNotInDatasetErrors or InvalidIdError for
        all update components that have invalid ids

    """
    errors = (
        _ids_valid_in_update_data_set(update_data, input_data, component, DatasetType.update)
        for component in update_data
    )
    return list(chain(*errors))


def validate_required_values(  # noqa: PLR0915
    data: SingleDataset, calculation_type: CalculationType | None = None, symmetric: bool = True
) -> list[MissingValueError]:
    """
    Checks if all required data is available.

    Args:
        data: A power-grid-model input dataset
        calculation_type: Supply a calculation method, to allow missing values for unused fields
        symmetric: A boolean to state whether input data will be used for a symmetric or asymmetric calculation

    Returns:
        An empty list if all required data is available, or a list of MissingValueErrors.
    """
    # Base
    required: dict[ComponentType | str, list[str]] = {"base": ["id"]}

    # Nodes
    required[ComponentType.node] = required["base"] + ["u_rated"]

    # Branches
    required["branch"] = required["base"] + ["from_node", "to_node", "from_status", "to_status"]
    required[ComponentType.link] = required["branch"].copy()
    required[ComponentType.line] = required["branch"] + ["r1", "x1", "c1", "tan1"]
    required[ComponentType.asym_line] = required["branch"] + [
        "r_aa",
        "r_ba",
        "r_bb",
        "r_ca",
        "r_cb",
        "r_cc",
        "x_aa",
        "x_ba",
        "x_bb",
        "x_ca",
        "x_cb",
        "x_cc",
    ]
    required[ComponentType.transformer] = required["branch"] + [
        "u1",
        "u2",
        "sn",
        "uk",
        "pk",
        "i0",
        "p0",
        "winding_from",
        "winding_to",
        "clock",
        "tap_side",
        "tap_min",
        "tap_max",
        "tap_size",
    ]
    # Branch3
    required["branch3"] = required["base"] + ["node_1", "node_2", "node_3", "status_1", "status_2", "status_3"]
    required[ComponentType.three_winding_transformer] = required["branch3"] + [
        "u1",
        "u2",
        "u3",
        "sn_1",
        "sn_2",
        "sn_3",
        "uk_12",
        "uk_13",
        "uk_23",
        "pk_12",
        "pk_13",
        "pk_23",
        "i0",
        "p0",
        "winding_1",
        "winding_2",
        "winding_3",
        "clock_12",
        "clock_13",
        "tap_side",
        "tap_min",
        "tap_max",
        "tap_size",
    ]

    # Regulators
    required["regulator"] = required["base"] + ["regulated_object", "status"]
    required[ComponentType.transformer_tap_regulator] = required["regulator"]
    if calculation_type is None or calculation_type == CalculationType.power_flow:
        required[ComponentType.transformer_tap_regulator] += ["control_side", "u_set", "u_band"]

    # Appliances
    required["appliance"] = required["base"] + ["node", "status"]
    required[ComponentType.source] = required["appliance"].copy()
    if calculation_type is None or calculation_type == CalculationType.power_flow:
        required[ComponentType.source] += ["u_ref"]
    required[ComponentType.shunt] = required["appliance"] + ["g1", "b1"]
    required["generic_load_gen"] = required["appliance"] + ["type"]
    if calculation_type is None or calculation_type == CalculationType.power_flow:
        required["generic_load_gen"] += ["p_specified", "q_specified"]
    required[ComponentType.sym_load] = required["generic_load_gen"].copy()
    required[ComponentType.asym_load] = required["generic_load_gen"].copy()
    required[ComponentType.sym_gen] = required["generic_load_gen"].copy()
    required[ComponentType.asym_gen] = required["generic_load_gen"].copy()

    # Sensors
    required["sensor"] = required["base"] + ["measured_object"]
    required["voltage_sensor"] = required["sensor"].copy()
    required["power_sensor"] = required["sensor"] + ["measured_terminal_type"]
    required["current_sensor"] = required["sensor"] + ["measured_terminal_type", "angle_measurement_type"]
    if calculation_type is None or calculation_type == CalculationType.state_estimation:
        required["voltage_sensor"] += ["u_sigma", "u_measured"]
        required["power_sensor"] += ["p_measured", "q_measured"]  # power_sigma, p_sigma and q_sigma are checked later
        required["current_sensor"] += ["i_sigma", "i_angle_sigma", "i_measured", "i_angle_measured"]
    required[ComponentType.sym_voltage_sensor] = required["voltage_sensor"].copy()
    required[ComponentType.asym_voltage_sensor] = required["voltage_sensor"].copy()
    required[ComponentType.sym_current_sensor] = required["current_sensor"].copy()
    required[ComponentType.asym_current_sensor] = required["current_sensor"].copy()

    # Different requirements for individual sensors. Avoid shallow copy.
    for sensor_type in (ComponentType.sym_power_sensor, ComponentType.asym_power_sensor):
        required[sensor_type] = required["power_sensor"].copy()

    # Faults
    required[ComponentType.fault] = required["base"] + ["fault_object"]
    asym_sc = False
    if calculation_type is None or calculation_type == CalculationType.short_circuit:
        required[ComponentType.fault] += ["status", "fault_type"]
        if ComponentType.fault in data:
            for elem in data[ComponentType.fault]["fault_type"]:
                if elem not in (FaultType.three_phase, FaultType.nan):
                    asym_sc = True
                    break

    if not symmetric or asym_sc:
        required[ComponentType.line] += ["r0", "x0", "c0", "tan0"]
        required[ComponentType.shunt] += ["g0", "b0"]

    errors = _validate_required_in_data(data, required)

    if calculation_type is None or calculation_type == CalculationType.state_estimation:
        errors += _validate_required_power_sigma_or_p_q_sigma(data, ComponentType.sym_power_sensor)
        errors += _validate_required_power_sigma_or_p_q_sigma(data, ComponentType.asym_power_sensor)

    return errors


def _validate_required_in_data(data: SingleDataset, required: dict[ComponentType | str, list[str]]):
    """
    Checks if all required data is available.

    Args:
        data: A power-grid-model input dataset
        required: a list of required fields (a list of str), per component when applicaple (a list of str or str lists)

    Returns:
        An empty list if all required data is available, or a list of MissingValueErrors.
    """

    def is_valid_component(data, component):
        return (
            not (isinstance(data[component], np.ndarray) and data[component].size == 0)
            and data[component] is not None
            and isinstance(data[component], ABCSized)
        )

    results: list[MissingValueError] = []

    for component in data:
        if is_valid_component(data, component):
            items = required.get(component, [])
            results += _none_missing(data, component, items)

    return results


def _validate_required_power_sigma_or_p_q_sigma(
    data: SingleDataset,
    power_sensor: Literal[ComponentType.sym_power_sensor, ComponentType.asym_power_sensor],
) -> list[MissingValueError]:
    """
    Check that either `p_sigma` and `q_sigma` are all provided, or that `power_sigma` is provided.

    Args:
        data: SingleDataset, pgm data
        sensor: the power sensor type, either ComponentType.sym_power_sensor or ComponentType.asym_power_sensor
    """
    result: list[MissingValueError] = []

    if power_sensor in data:
        sensor_data = data[power_sensor]
        p_sigma = sensor_data["p_sigma"]
        q_sigma = sensor_data["q_sigma"]

        asym_axes = tuple(range(sensor_data.ndim, p_sigma.ndim))
        all_pq_sigma_missing_mask = np.all(np.isnan(p_sigma), axis=asym_axes) & np.all(
            np.isnan(q_sigma), axis=asym_axes
        )

        result += _validate_required_in_data(
            {power_sensor: sensor_data[all_pq_sigma_missing_mask]}, required={power_sensor: ["power_sigma"]}
        )
        result += _validate_required_in_data(
            {power_sensor: sensor_data[~all_pq_sigma_missing_mask]}, required={power_sensor: ["p_sigma", "q_sigma"]}
        )

    return result


def validate_values(data: SingleDataset, calculation_type: CalculationType | None = None) -> list[ValidationError]:
    """
    For each component supplied in the data, call the appropriate validation function

    Args:
        data: A power-grid-model input dataset
        calculation_type: Supply a calculation method, to allow missing values for unused fields

    Returns:
        An empty list if all required data is valid, or a list of ValidationErrors.

    """
    errors: list[ValidationError] = list(
        _all_finite(
            data=data,
            exceptions={
                ComponentType.sym_power_sensor: ["power_sigma", "p_sigma", "q_sigma"],
                ComponentType.asym_power_sensor: ["power_sigma", "p_sigma", "q_sigma"],
                ComponentType.sym_voltage_sensor: ["u_sigma"],
                ComponentType.asym_voltage_sensor: ["u_sigma"],
                ComponentType.sym_current_sensor: ["i_sigma", "i_angle_sigma"],
                ComponentType.asym_current_sensor: ["i_sigma", "i_angle_sigma"],
            },
        )
    )

    component_validators = {
        ComponentType.node: validate_node,
        ComponentType.line: validate_line,
        ComponentType.asym_line: validate_asym_line,
        ComponentType.link: lambda d: validate_branch(d, ComponentType.link),
        ComponentType.generic_branch: validate_generic_branch,
        ComponentType.transformer: validate_transformer,
        ComponentType.three_winding_transformer: validate_three_winding_transformer,
        ComponentType.source: validate_source,
        ComponentType.sym_load: lambda d: validate_generic_load_gen(d, ComponentType.sym_load),
        ComponentType.sym_gen: lambda d: validate_generic_load_gen(d, ComponentType.sym_gen),
        ComponentType.asym_load: lambda d: validate_generic_load_gen(d, ComponentType.asym_load),
        ComponentType.asym_gen: lambda d: validate_generic_load_gen(d, ComponentType.asym_gen),
        ComponentType.shunt: validate_shunt,
    }

    for component, validator in component_validators.items():
        if component in data:
            errors += validator(data)

    if calculation_type in (None, CalculationType.state_estimation):
        if ComponentType.sym_voltage_sensor in data:
            errors += validate_generic_voltage_sensor(data, ComponentType.sym_voltage_sensor)
        if ComponentType.asym_voltage_sensor in data:
            errors += validate_generic_voltage_sensor(data, ComponentType.asym_voltage_sensor)
        if ComponentType.sym_power_sensor in data:
            errors += validate_generic_power_sensor(data, ComponentType.sym_power_sensor)
        if ComponentType.asym_power_sensor in data:
            errors += validate_generic_power_sensor(data, ComponentType.asym_power_sensor)
        if ComponentType.sym_current_sensor in data:
            errors += validate_generic_current_sensor(data, ComponentType.sym_current_sensor)
        if ComponentType.asym_current_sensor in data:
            errors += validate_generic_current_sensor(data, ComponentType.asym_current_sensor)

        errors += validate_no_mixed_sensors_on_same_terminal(data)

    if calculation_type in (None, CalculationType.short_circuit) and ComponentType.fault in data:
        errors += validate_fault(data)

    if calculation_type in (None, CalculationType.power_flow) and ComponentType.transformer_tap_regulator in data:
        errors += validate_transformer_tap_regulator(data)

    return errors


def validate_base(data: SingleDataset, component: ComponentType) -> list[ValidationError]:
    errors: list[ValidationError] = list(_all_unique(data, component, "id"))
    return errors


def validate_node(data: SingleDataset) -> list[ValidationError]:
    errors = validate_base(data, ComponentType.node)
    errors += _all_greater_than_zero(data, ComponentType.node, "u_rated")
    return errors


def validate_branch(data: SingleDataset, component: ComponentType) -> list[ValidationError]:
    errors = validate_base(data, component)
    errors += _all_valid_ids(data, component, "from_node", ComponentType.node)
    errors += _all_valid_ids(data, component, "to_node", ComponentType.node)
    errors += _all_not_two_values_equal(data, component, "to_node", "from_node")
    errors += _all_boolean(data, component, "from_status")
    errors += _all_boolean(data, component, "to_status")
    return errors


def validate_line(data: SingleDataset) -> list[ValidationError]:
    errors = validate_branch(data, ComponentType.line)
    errors += _all_not_two_values_zero(data, ComponentType.line, "r1", "x1")
    errors += _all_not_two_values_zero(data, ComponentType.line, "r0", "x0")
    errors += _all_greater_than_zero(data, ComponentType.line, "i_n")
    return errors


def validate_asym_line(data: SingleDataset) -> list[ValidationError]:
    errors = validate_branch(data, ComponentType.asym_line)
    errors += _all_greater_than_zero(data, ComponentType.asym_line, "i_n")
    required_fields = ["r_aa", "r_ba", "r_bb", "r_ca", "r_cb", "r_cc", "x_aa", "x_ba", "x_bb", "x_ca", "x_cb", "x_cc"]
    optional_r_matrix_fields = ["r_na", "r_nb", "r_nc", "r_nn"]
    optional_x_matrix_fields = ["x_na", "x_nb", "x_nc", "x_nn"]
    required_c_matrix_fields = ["c_aa", "c_ba", "c_bb", "c_ca", "c_cb", "c_cc"]
    c_fields = ["c0", "c1"]
    for field in (
        required_fields + optional_r_matrix_fields + optional_x_matrix_fields + required_c_matrix_fields + c_fields
    ):
        errors += _all_greater_than_zero(data, ComponentType.asym_line, field)

    errors += _no_strict_subset_missing(
        data, optional_r_matrix_fields + optional_x_matrix_fields, ComponentType.asym_line
    )
    errors += _no_strict_subset_missing(data, required_c_matrix_fields, ComponentType.asym_line)
    errors += _no_strict_subset_missing(data, c_fields, ComponentType.asym_line)
    errors += _not_all_missing(data, required_c_matrix_fields + c_fields, ComponentType.asym_line)

    return errors


def validate_generic_branch(data: SingleDataset) -> list[ValidationError]:
    errors = validate_branch(data, ComponentType.generic_branch)
    errors += _all_greater_than_zero(data, ComponentType.generic_branch, "k")
    errors += _all_greater_than_or_equal_to_zero(data, ComponentType.generic_branch, "sn")
    return errors


def validate_transformer(data: SingleDataset) -> list[ValidationError]:
    errors = validate_branch(data, ComponentType.transformer)
    errors += _all_greater_than_zero(data, ComponentType.transformer, "u1")
    errors += _all_greater_than_zero(data, ComponentType.transformer, "u2")
    errors += _all_greater_than_zero(data, ComponentType.transformer, "sn")
    errors += _all_greater_or_equal(data, ComponentType.transformer, "uk", "pk/sn")
    errors += _all_between(data, ComponentType.transformer, "uk", 0, 1)
    errors += _all_greater_than_or_equal_to_zero(data, ComponentType.transformer, "pk")
    errors += _all_greater_or_equal(data, ComponentType.transformer, "i0", "p0/sn")
    errors += _all_less_than(data, ComponentType.transformer, "i0", 1)
    errors += _all_greater_than_or_equal_to_zero(data, ComponentType.transformer, "p0")
    errors += _all_valid_enum_values(data, ComponentType.transformer, "winding_from", WindingType)
    errors += _all_valid_enum_values(data, ComponentType.transformer, "winding_to", WindingType)
    errors += _all_between_or_at(data, ComponentType.transformer, "clock", -12, 12)
    errors += _all_valid_clocks(data, ComponentType.transformer, "clock", "winding_from", "winding_to")
    errors += _all_valid_enum_values(data, ComponentType.transformer, "tap_side", BranchSide)
    errors += _all_between_or_at(
        data, ComponentType.transformer, "tap_pos", "tap_min", "tap_max", data[ComponentType.transformer]["tap_nom"], 0
    )
    errors += _all_between_or_at(data, ComponentType.transformer, "tap_nom", "tap_min", "tap_max", 0)
    errors += _all_greater_than_or_equal_to_zero(data, ComponentType.transformer, "tap_size")
    errors += _all_greater_or_equal(
        data, ComponentType.transformer, "uk_min", "pk_min/sn", data[ComponentType.transformer]["uk"]
    )
    errors += _all_between(data, ComponentType.transformer, "uk_min", 0, 1, data[ComponentType.transformer]["uk"])
    errors += _all_greater_or_equal(
        data, ComponentType.transformer, "uk_max", "pk_max/sn", data[ComponentType.transformer]["uk"]
    )
    errors += _all_between(data, ComponentType.transformer, "uk_max", 0, 1, data[ComponentType.transformer]["uk"])
    errors += _all_greater_than_or_equal_to_zero(
        data, ComponentType.transformer, "pk_min", data[ComponentType.transformer]["pk"]
    )
    errors += _all_greater_than_or_equal_to_zero(
        data, ComponentType.transformer, "pk_max", data[ComponentType.transformer]["pk"]
    )
    return errors


def validate_branch3(data: SingleDataset, component: ComponentType) -> list[ValidationError]:
    errors = validate_base(data, component)
    errors += _all_valid_ids(data, component, "node_1", ComponentType.node)
    errors += _all_valid_ids(data, component, "node_2", ComponentType.node)
    errors += _all_valid_ids(data, component, "node_3", ComponentType.node)
    errors += _all_not_two_values_equal(data, component, "node_1", "node_2")
    errors += _all_not_two_values_equal(data, component, "node_1", "node_3")
    errors += _all_not_two_values_equal(data, component, "node_2", "node_3")
    errors += _all_boolean(data, component, "status_1")
    errors += _all_boolean(data, component, "status_2")
    errors += _all_boolean(data, component, "status_3")
    return errors


def validate_three_winding_transformer(data: SingleDataset) -> list[ValidationError]:  # noqa: PLR0915
    errors = validate_branch3(data, ComponentType.three_winding_transformer)
    errors += _all_greater_than_zero(data, ComponentType.three_winding_transformer, "u1")
    errors += _all_greater_than_zero(data, ComponentType.three_winding_transformer, "u2")
    errors += _all_greater_than_zero(data, ComponentType.three_winding_transformer, "u3")
    errors += _all_greater_than_zero(data, ComponentType.three_winding_transformer, "sn_1")
    errors += _all_greater_than_zero(data, ComponentType.three_winding_transformer, "sn_2")
    errors += _all_greater_than_zero(data, ComponentType.three_winding_transformer, "sn_3")
    errors += _all_greater_or_equal(data, ComponentType.three_winding_transformer, "uk_12", "pk_12/sn_1")
    errors += _all_greater_or_equal(data, ComponentType.three_winding_transformer, "uk_12", "pk_12/sn_2")
    errors += _all_greater_or_equal(data, ComponentType.three_winding_transformer, "uk_13", "pk_13/sn_1")
    errors += _all_greater_or_equal(data, ComponentType.three_winding_transformer, "uk_13", "pk_13/sn_3")
    errors += _all_greater_or_equal(data, ComponentType.three_winding_transformer, "uk_23", "pk_23/sn_2")
    errors += _all_greater_or_equal(data, ComponentType.three_winding_transformer, "uk_23", "pk_23/sn_3")
    errors += _all_between(data, ComponentType.three_winding_transformer, "uk_12", 0, 1)
    errors += _all_between(data, ComponentType.three_winding_transformer, "uk_13", 0, 1)
    errors += _all_between(data, ComponentType.three_winding_transformer, "uk_23", 0, 1)
    errors += _all_greater_than_or_equal_to_zero(data, ComponentType.three_winding_transformer, "pk_12")
    errors += _all_greater_than_or_equal_to_zero(data, ComponentType.three_winding_transformer, "pk_13")
    errors += _all_greater_than_or_equal_to_zero(data, ComponentType.three_winding_transformer, "pk_23")
    errors += _all_greater_or_equal(data, ComponentType.three_winding_transformer, "i0", "p0/sn_1")
    errors += _all_less_than(data, ComponentType.three_winding_transformer, "i0", 1)
    errors += _all_greater_than_or_equal_to_zero(data, ComponentType.three_winding_transformer, "p0")
    errors += _all_valid_enum_values(data, ComponentType.three_winding_transformer, "winding_1", WindingType)
    errors += _all_valid_enum_values(data, ComponentType.three_winding_transformer, "winding_2", WindingType)
    errors += _all_valid_enum_values(data, ComponentType.three_winding_transformer, "winding_3", WindingType)
    errors += _all_between_or_at(data, ComponentType.three_winding_transformer, "clock_12", -12, 12)
    errors += _all_between_or_at(data, ComponentType.three_winding_transformer, "clock_13", -12, 12)
    errors += _all_valid_clocks(data, ComponentType.three_winding_transformer, "clock_12", "winding_1", "winding_2")
    errors += _all_valid_clocks(data, ComponentType.three_winding_transformer, "clock_13", "winding_1", "winding_3")
    errors += _all_valid_enum_values(data, ComponentType.three_winding_transformer, "tap_side", Branch3Side)
    errors += _all_between_or_at(
        data,
        ComponentType.three_winding_transformer,
        "tap_pos",
        "tap_min",
        "tap_max",
        data[ComponentType.three_winding_transformer]["tap_nom"],
        0,
    )
    errors += _all_between_or_at(data, ComponentType.three_winding_transformer, "tap_nom", "tap_min", "tap_max", 0)
    errors += _all_greater_than_or_equal_to_zero(data, ComponentType.three_winding_transformer, "tap_size")
    errors += _all_greater_or_equal(
        data,
        ComponentType.three_winding_transformer,
        "uk_12_min",
        "pk_12_min/sn_1",
        data[ComponentType.three_winding_transformer]["uk_12"],
    )
    errors += _all_greater_or_equal(
        data,
        ComponentType.three_winding_transformer,
        "uk_12_min",
        "pk_12_min/sn_2",
        data[ComponentType.three_winding_transformer]["uk_12"],
    )
    errors += _all_greater_or_equal(
        data,
        ComponentType.three_winding_transformer,
        "uk_13_min",
        "pk_13_min/sn_1",
        data[ComponentType.three_winding_transformer]["uk_13"],
    )
    errors += _all_greater_or_equal(
        data,
        ComponentType.three_winding_transformer,
        "uk_13_min",
        "pk_13_min/sn_3",
        data[ComponentType.three_winding_transformer]["uk_13"],
    )
    errors += _all_greater_or_equal(
        data,
        ComponentType.three_winding_transformer,
        "uk_23_min",
        "pk_23_min/sn_2",
        data[ComponentType.three_winding_transformer]["uk_23"],
    )
    errors += _all_greater_or_equal(
        data,
        ComponentType.three_winding_transformer,
        "uk_23_min",
        "pk_23_min/sn_3",
        data[ComponentType.three_winding_transformer]["uk_23"],
    )
    errors += _all_between(
        data,
        ComponentType.three_winding_transformer,
        "uk_12_min",
        0,
        1,
        data[ComponentType.three_winding_transformer]["uk_12"],
    )
    errors += _all_between(
        data,
        ComponentType.three_winding_transformer,
        "uk_13_min",
        0,
        1,
        data[ComponentType.three_winding_transformer]["uk_13"],
    )
    errors += _all_between(
        data,
        ComponentType.three_winding_transformer,
        "uk_23_min",
        0,
        1,
        data[ComponentType.three_winding_transformer]["uk_23"],
    )
    errors += _all_greater_or_equal(
        data,
        ComponentType.three_winding_transformer,
        "uk_12_max",
        "pk_12_max/sn_1",
        data[ComponentType.three_winding_transformer]["uk_12"],
    )
    errors += _all_greater_or_equal(
        data,
        ComponentType.three_winding_transformer,
        "uk_12_max",
        "pk_12_max/sn_2",
        data[ComponentType.three_winding_transformer]["uk_12"],
    )
    errors += _all_greater_or_equal(
        data,
        ComponentType.three_winding_transformer,
        "uk_13_max",
        "pk_13_max/sn_1",
        data[ComponentType.three_winding_transformer]["uk_13"],
    )
    errors += _all_greater_or_equal(
        data,
        ComponentType.three_winding_transformer,
        "uk_13_max",
        "pk_13_max/sn_3",
        data[ComponentType.three_winding_transformer]["uk_13"],
    )
    errors += _all_greater_or_equal(
        data,
        ComponentType.three_winding_transformer,
        "uk_23_max",
        "pk_23_max/sn_2",
        data[ComponentType.three_winding_transformer]["uk_23"],
    )
    errors += _all_greater_or_equal(
        data,
        ComponentType.three_winding_transformer,
        "uk_23_max",
        "pk_23_max/sn_3",
        data[ComponentType.three_winding_transformer]["uk_23"],
    )
    errors += _all_between(
        data,
        ComponentType.three_winding_transformer,
        "uk_12_max",
        0,
        1,
        data[ComponentType.three_winding_transformer]["uk_12"],
    )
    errors += _all_between(
        data,
        ComponentType.three_winding_transformer,
        "uk_13_max",
        0,
        1,
        data[ComponentType.three_winding_transformer]["uk_13"],
    )
    errors += _all_between(
        data,
        ComponentType.three_winding_transformer,
        "uk_23_max",
        0,
        1,
        data[ComponentType.three_winding_transformer]["uk_23"],
    )
    errors += _all_greater_than_or_equal_to_zero(
        data,
        ComponentType.three_winding_transformer,
        "pk_12_min",
        data[ComponentType.three_winding_transformer]["pk_12"],
    )
    errors += _all_greater_than_or_equal_to_zero(
        data,
        ComponentType.three_winding_transformer,
        "pk_13_min",
        data[ComponentType.three_winding_transformer]["pk_13"],
    )
    errors += _all_greater_than_or_equal_to_zero(
        data,
        ComponentType.three_winding_transformer,
        "pk_23_min",
        data[ComponentType.three_winding_transformer]["pk_23"],
    )
    errors += _all_greater_than_or_equal_to_zero(
        data,
        ComponentType.three_winding_transformer,
        "pk_12_max",
        data[ComponentType.three_winding_transformer]["pk_12"],
    )
    errors += _all_greater_than_or_equal_to_zero(
        data,
        ComponentType.three_winding_transformer,
        "pk_13_max",
        data[ComponentType.three_winding_transformer]["pk_13"],
    )
    errors += _all_greater_than_or_equal_to_zero(
        data,
        ComponentType.three_winding_transformer,
        "pk_23_max",
        data[ComponentType.three_winding_transformer]["pk_23"],
    )
    return errors


def validate_appliance(data: SingleDataset, component: ComponentType) -> list[ValidationError]:
    errors = validate_base(data, component)
    errors += _all_boolean(data, component, "status")
    errors += _all_valid_ids(data, component, "node", ComponentType.node)
    return errors


def validate_source(data: SingleDataset) -> list[ValidationError]:
    errors = validate_appliance(data, ComponentType.source)
    errors += _all_greater_than_zero(data, ComponentType.source, "u_ref")
    errors += _all_greater_than_zero(data, ComponentType.source, "sk")
    errors += _all_greater_than_or_equal_to_zero(data, ComponentType.source, "rx_ratio")
    errors += _all_greater_than_zero(data, ComponentType.source, "z01_ratio")
    return errors


def validate_generic_load_gen(data: SingleDataset, component: ComponentType) -> list[ValidationError]:
    errors = validate_appliance(data, component)
    errors += _all_valid_enum_values(data, component, "type", LoadGenType)
    return errors


def validate_shunt(data: SingleDataset) -> list[ValidationError]:
    return validate_appliance(data, ComponentType.shunt)


def validate_generic_voltage_sensor(data: SingleDataset, component: ComponentType) -> list[ValidationError]:
    errors = validate_base(data, component)
    errors += _all_greater_than_zero(data, component, "u_sigma")
    errors += _all_greater_than_zero(data, component, "u_measured")
    errors += _all_valid_ids(data, component, "measured_object", ComponentType.node)
    return errors


def validate_generic_power_sensor(data: SingleDataset, component: ComponentType) -> list[ValidationError]:
    errors = validate_base(data, component)
    errors += _all_greater_than_zero(data, component, "power_sigma")
    errors += _all_valid_enum_values(data, component, "measured_terminal_type", MeasuredTerminalType)
    errors += _all_valid_ids(
        data,
        component,
        field="measured_object",
        ref_components=[
            ComponentType.node,
            ComponentType.line,
            ComponentType.asym_line,
            ComponentType.generic_branch,
            ComponentType.transformer,
            ComponentType.three_winding_transformer,
            ComponentType.source,
            ComponentType.shunt,
            ComponentType.sym_load,
            ComponentType.asym_load,
            ComponentType.sym_gen,
            ComponentType.asym_gen,
        ],
    )
    errors += _all_valid_ids(
        data,
        component,
        field="measured_object",
        ref_components=[
            ComponentType.line,
            ComponentType.asym_line,
            ComponentType.generic_branch,
            ComponentType.transformer,
        ],
        measured_terminal_type=MeasuredTerminalType.branch_from,
    )
    errors += _all_valid_ids(
        data,
        component,
        field="measured_object",
        ref_components=[
            ComponentType.line,
            ComponentType.asym_line,
            ComponentType.generic_branch,
            ComponentType.transformer,
        ],
        measured_terminal_type=MeasuredTerminalType.branch_to,
    )
    errors += _all_valid_ids(
        data,
        component,
        field="measured_object",
        ref_components=ComponentType.source,
        measured_terminal_type=MeasuredTerminalType.source,
    )
    errors += _all_valid_ids(
        data,
        component,
        field="measured_object",
        ref_components=ComponentType.shunt,
        measured_terminal_type=MeasuredTerminalType.shunt,
    )
    errors += _all_valid_ids(
        data,
        component,
        field="measured_object",
        ref_components=[ComponentType.sym_load, ComponentType.asym_load],
        measured_terminal_type=MeasuredTerminalType.load,
    )
    errors += _all_valid_ids(
        data,
        component,
        field="measured_object",
        ref_components=[ComponentType.sym_gen, ComponentType.asym_gen],
        measured_terminal_type=MeasuredTerminalType.generator,
    )
    errors += _all_valid_ids(
        data,
        component,
        field="measured_object",
        ref_components=ComponentType.three_winding_transformer,
        measured_terminal_type=MeasuredTerminalType.branch3_1,
    )
    errors += _all_valid_ids(
        data,
        component,
        field="measured_object",
        ref_components=ComponentType.three_winding_transformer,
        measured_terminal_type=MeasuredTerminalType.branch3_2,
    )
    errors += _all_valid_ids(
        data,
        component,
        field="measured_object",
        ref_components=ComponentType.three_winding_transformer,
        measured_terminal_type=MeasuredTerminalType.branch3_3,
    )
    errors += _all_valid_ids(
        data,
        component,
        field="measured_object",
        ref_components=ComponentType.node,
        measured_terminal_type=MeasuredTerminalType.node,
    )
    if component in (ComponentType.sym_power_sensor, ComponentType.asym_power_sensor):
        errors += _valid_p_q_sigma(data, component)

    return errors


def validate_generic_current_sensor(data: SingleDataset, component: ComponentType) -> list[ValidationError]:
    errors = validate_base(data, component)
    errors += _all_greater_than_zero(data, component, "i_sigma")
    errors += _all_greater_than_zero(data, component, "i_angle_sigma")
    errors += _all_valid_enum_values(data, component, "measured_terminal_type", MeasuredTerminalType)
    errors += _all_valid_enum_values(data, component, "angle_measurement_type", AngleMeasurementType)
    errors += _all_in_valid_values(
        data,
        component,
        "measured_terminal_type",
        [
            MeasuredTerminalType.branch_from,
            MeasuredTerminalType.branch_to,
            MeasuredTerminalType.branch3_1,
            MeasuredTerminalType.branch3_2,
            MeasuredTerminalType.branch3_3,
        ],
    )
    errors += _all_valid_ids(
        data,
        component,
        field="measured_object",
        ref_components=[
            ComponentType.line,
            ComponentType.asym_line,
            ComponentType.generic_branch,
            ComponentType.transformer,
            ComponentType.three_winding_transformer,
        ],
    )
    errors += _all_valid_ids(
        data,
        component,
        field="measured_object",
        ref_components=[
            ComponentType.line,
            ComponentType.asym_line,
            ComponentType.generic_branch,
            ComponentType.transformer,
        ],
        measured_terminal_type=MeasuredTerminalType.branch_from,
    )
    errors += _all_valid_ids(
        data,
        component,
        field="measured_object",
        ref_components=[
            ComponentType.line,
            ComponentType.asym_line,
            ComponentType.generic_branch,
            ComponentType.transformer,
        ],
        measured_terminal_type=MeasuredTerminalType.branch_to,
    )
    errors += _all_valid_ids(
        data,
        component,
        field="measured_object",
        ref_components=ComponentType.three_winding_transformer,
        measured_terminal_type=MeasuredTerminalType.branch3_1,
    )
    errors += _all_valid_ids(
        data,
        component,
        field="measured_object",
        ref_components=ComponentType.three_winding_transformer,
        measured_terminal_type=MeasuredTerminalType.branch3_2,
    )
    errors += _all_valid_ids(
        data,
        component,
        field="measured_object",
        ref_components=ComponentType.three_winding_transformer,
        measured_terminal_type=MeasuredTerminalType.branch3_3,
    )
    errors += _all_same_current_angle_measurement_type_on_terminal(
        data,
        component,
        measured_object_field="measured_object",
        measured_terminal_type_field="measured_terminal_type",
        angle_measurement_type_field="angle_measurement_type",
    )
    errors += _any_voltage_angle_measurement_if_global_current_measurement(
        data,
        component,
        angle_measurement_type_filter=("angle_measurement_type", AngleMeasurementType.global_angle),
        voltage_sensor_u_angle_measured={
            ComponentType.sym_voltage_sensor: "u_angle_measured",
            ComponentType.asym_voltage_sensor: "u_angle_measured",
        },
    )

    return errors


def validate_fault(data: SingleDataset) -> list[ValidationError]:
    errors = validate_base(data, ComponentType.fault)
    errors += _all_boolean(data, ComponentType.fault, "status")
    errors += _all_valid_enum_values(data, ComponentType.fault, "fault_type", FaultType)
    errors += _all_valid_enum_values(data, ComponentType.fault, "fault_phase", FaultPhase)
    errors += _all_valid_fault_phases(data, ComponentType.fault, "fault_type", "fault_phase")
    errors += _all_valid_ids(data, ComponentType.fault, field="fault_object", ref_components=ComponentType.node)
    errors += _all_greater_than_or_equal_to_zero(data, ComponentType.fault, "r_f")
    errors += _all_enabled_identical(data, ComponentType.fault, "fault_type", "status")
    errors += _all_enabled_identical(data, ComponentType.fault, "fault_phase", "status")
    return errors


def validate_regulator(data: SingleDataset, component: ComponentType) -> list[ValidationError]:
    errors = validate_base(data, component)
    errors += _all_valid_ids(
        data,
        component,
        field="regulated_object",
        ref_components=[ComponentType.transformer, ComponentType.three_winding_transformer],
    )
    return errors


def validate_transformer_tap_regulator(data: SingleDataset) -> list[ValidationError]:
    errors = validate_regulator(data, ComponentType.transformer_tap_regulator)
    errors += _all_boolean(data, ComponentType.transformer_tap_regulator, "status")
    errors += _all_unique(data, ComponentType.transformer_tap_regulator, "regulated_object")
    errors += _all_valid_enum_values(
        data, ComponentType.transformer_tap_regulator, "control_side", [BranchSide, Branch3Side]
    )
    errors += _all_valid_associated_enum_values(
        data,
        ComponentType.transformer_tap_regulator,
        "control_side",
        "regulated_object",
        [ComponentType.transformer],
        [BranchSide],
    )
    errors += _all_valid_associated_enum_values(
        data,
        ComponentType.transformer_tap_regulator,
        "control_side",
        "regulated_object",
        [ComponentType.three_winding_transformer],
        [Branch3Side],
    )
    errors += _all_greater_than_or_equal_to_zero(data, ComponentType.transformer_tap_regulator, "u_set")
    errors += _all_greater_than_zero(data, ComponentType.transformer_tap_regulator, "u_band")
    errors += _all_greater_than_or_equal_to_zero(
        data, ComponentType.transformer_tap_regulator, "line_drop_compensation_r", 0.0
    )
    errors += _all_greater_than_or_equal_to_zero(
        data, ComponentType.transformer_tap_regulator, "line_drop_compensation_x", 0.0
    )
    return errors


def validate_no_mixed_sensors_on_same_terminal(data: SingleDataset) -> list[ValidationError]:
    errors: list[ValidationError] = []

    for power_sensor in [ComponentType.sym_power_sensor, ComponentType.asym_power_sensor]:
        for current_sensor in [ComponentType.sym_current_sensor, ComponentType.asym_current_sensor]:
            if power_sensor in data and current_sensor in data:
                errors += _all_same_sensor_type_on_same_terminal(
                    data,
                    power_sensor_type=power_sensor,
                    current_sensor_type=current_sensor,
                    measured_object_field="measured_object",
                    measured_terminal_type_field="measured_terminal_type",
                )

    return errors



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

import json
import os
import re
from pathlib import Path
from typing import Any

import numpy as np
import pytest

from power_grid_model._core.dataset_definitions import DatasetType
from power_grid_model._core.power_grid_model import PowerGridModel
from power_grid_model.data_types import Dataset, PythonDataset, SingleDataset
from power_grid_model.errors import (
    AutomaticTapCalculationError,
    AutomaticTapInputError,
    ConflictID,
    ConflictVoltage,
    IDNotFound,
    IDWrongType,
    InvalidBranch,
    InvalidBranch3,
    InvalidCalculationMethod,
    InvalidMeasuredObject,
    InvalidRegulatedObject,
    InvalidTransformerClock,
    MaxIterationReached,
    NotObservableError,
    PowerGridBatchError,
    PowerGridError,
    PowerGridSerializationError,
    SparseMatrixError,
)
from power_grid_model.utils import json_deserialize, json_deserialize_from_file, json_serialize_to_file

_Failed: type[BaseException] | None
try:
    from _pytest.outcomes import Failed as _PytestFailedOutcome

    _Failed = _PytestFailedOutcome
except ImportError:
    import warnings

    warnings.warn(
        """Failed to import _pytest.outcomes.Failed."""
        """ Some validation cases tests marked as {"xfail": {"raises": "Failed"}} may report as xfail"""
        """ even though they are actual problems."""
    )
    _Failed = None


BASE_PATH = Path(__file__).parent.parent
DATA_PATH = BASE_PATH / "data"
OUTPUT_PATH = BASE_PATH / "output"
EXPORT_OUTPUT = ("POWER_GRID_MODEL_VALIDATION_TEST_EXPORT" in os.environ) and (
    os.environ["POWER_GRID_MODEL_VALIDATION_TEST_EXPORT"] == "ON"
)

KNOWN_EXCEPTIONS: dict[str, type[BaseException] | None] = {
    ex.__name__: ex
    for ex in (
        PowerGridBatchError,
        PowerGridError,
        ConflictID,
        ConflictVoltage,
        IDNotFound,
        IDWrongType,
        InvalidBranch,
        InvalidBranch3,
        InvalidCalculationMethod,
        InvalidMeasuredObject,
        InvalidRegulatedObject,
        AutomaticTapCalculationError,
        AutomaticTapInputError,
        InvalidTransformerClock,
        NotObservableError,
        SparseMatrixError,
        PowerGridSerializationError,
        AssertionError,
        OSError,
        MaxIterationReached,
    )
}
KNOWN_EXCEPTIONS["Failed"] = _Failed


class UnknownDataType(Exception):
    """Raised when an unknown data type is encountered"""


class PowerGridModelWithExt(PowerGridModel):
    """Wrapper class around the power grid model to expose extended features."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def calculate_power_flow_with_ext(self, *args, **kwargs):
        """calculate_power_flow with extended features."""
        return self._calculate_power_flow(*args, **kwargs)

    def calculate_state_estimation_with_ext(self, *args, **kwargs):
        """calculate_state_estimation with extended features."""
        return self._calculate_state_estimation(*args, **kwargs)

    def calculate_short_circuit_with_ext(self, *args, **kwargs):
        """calculate_short_circuit with extended features."""
        return self._calculate_short_circuit(*args, **kwargs)


def get_output_type(calculation_type: str, sym: bool) -> str:
    if calculation_type == "short_circuit":
        if sym:
            raise AssertionError(f"Unsupported validation case: calculation_type={calculation_type}, sym={sym}")
        return "sc_output"

    if sym:
        return "sym_output"

    return "asym_output"


def get_test_case_paths(calculation_type: str, test_cases: list[str] | None = None) -> dict[str, Path]:
    """get a list of all cases, directories in validation datasets"""
    calculation_type_dir = DATA_PATH / calculation_type
    test_case_paths = {
        str(item.relative_to(DATA_PATH)).replace("\\", "/"): item
        for item in calculation_type_dir.glob("**/")
        if (item.is_dir() and (item / "params.json").is_file())
    }
    if test_cases is not None:
        test_case_paths = {key: value for key, value in test_case_paths.items() if key in test_cases}

    return test_case_paths


def add_case(
    case_name: str,
    case_dir: Path,
    params: dict,
    calculation_type: str,
    is_batch: bool,
    calculation_method: str,
    sym: bool,
):
    output_prefix = get_output_type(calculation_type=calculation_type, sym=sym)
    batch_suffix = "_batch" if is_batch else ""

    # only generate a case if sym or asym output exists
    if (case_dir / f"{output_prefix}{batch_suffix}.json").exists():
        # Build a recognizable case ID
        case_id = case_name
        case_id += "-sym" if sym else "-asym"
        case_id += "-" + calculation_method
        if is_batch:
            case_id += "-batch"

        calculation_method_params = dict(params, **params.get("extra_params", {}).get(calculation_method, {}))
        pytest_param = [
            case_id,
            case_dir,
            sym,
            calculation_type,
            calculation_method,
            calculation_method_params["rtol"],
            calculation_method_params["atol"],
            calculation_method_params,
        ]
        kwargs = {}
        if "xfail" in calculation_method_params:
            xfail = calculation_method_params["xfail"]
            kwargs["marks"] = pytest.mark.xfail(
                reason=xfail["reason"], raises=KNOWN_EXCEPTIONS[xfail.get("raises", "AssertionError")]
            )

        yield pytest.param(*pytest_param, **kwargs, id=case_id)


def _add_cases(case_dir: Path, calculation_type: str, **kwargs):
    with (case_dir / "params.json").open() as f:
        params = json.load(f)

    # retrieve calculation method, can be a string or list of strings
    calculation_methods = params["calculation_method"]
    if not isinstance(calculation_methods, list):
        calculation_methods = [calculation_methods]

    # loop for sym or asym scenario
    for calculation_method in calculation_methods:
        for sym in [True, False]:
            if calculation_type == "short_circuit" and sym:
                continue  # only asym short circuit calculations are supported

            yield from add_case(
                case_dir=case_dir,
                params=params,
                calculation_type=calculation_type,
                calculation_method=calculation_method,
                sym=sym,
                **kwargs,
            )


def pytest_cases(get_batch_cases: bool = False, data_dir: str | None = None, test_cases: list[str] | None = None):
    relevant_calculations = [data_dir] if data_dir is not None else ["power_flow", "state_estimation", "short_circuit"]

    for calculation_type in relevant_calculations:
        test_case_paths = get_test_case_paths(calculation_type=calculation_type, test_cases=test_cases)

        for case_name, case_dir in test_case_paths.items():
            yield from _add_cases(
                case_name=case_name,
                case_dir=case_dir,
                calculation_type=calculation_type,
                is_batch=get_batch_cases,
            )


def bool_params(true_id: str, false_id: str | None = None, **kwargs):
    if false_id is None:
        false_id = f"not-{true_id}"
    yield pytest.param(False, **kwargs, id=false_id)
    yield pytest.param(True, **kwargs, id=true_id)


def dict_params(params: dict[Any, str], **kwargs):
    for value, param_id in params.items():
        yield pytest.param(value, **kwargs, id=param_id)


def import_case_data(data_path: Path, calculation_type: str, sym: bool):
    output_prefix = get_output_type(calculation_type=calculation_type, sym=sym)
    return_dict = {"input": json_deserialize_from_file(data_path / "input.json")}
    # import output if relevant
    if (data_path / f"{output_prefix}.json").exists():
        return_dict["output"] = json_deserialize_from_file(data_path / f"{output_prefix}.json")
    # import update and output batch if relevant
    if (data_path / "update_batch.json").exists():
        return_dict["update_batch"] = json_deserialize_from_file(data_path / "update_batch.json")
        return_dict["output_batch"] = json_deserialize_from_file(data_path / f"{output_prefix}_batch.json")

    return return_dict


def save_json_data(json_file: str, data: Dataset):
    OUTPUT_PATH.mkdir(parents=True, exist_ok=True)
    data_file = OUTPUT_PATH / json_file
    data_file.parent.mkdir(parents=True, exist_ok=True)
    json_serialize_to_file(data_file, data)


def compare_result(actual: SingleDataset, expected: SingleDataset, rtol: float, atol: float | dict[str, float]):
    actual_col_ndim = 2
    for key, expected_data in expected.items():
        if not isinstance(expected_data, np.ndarray):
            raise NotImplementedError("Validation tests are not implemented for columnar data")

        for col_name in expected_data.dtype.names:
            actual_col = actual[key][col_name]
            expected_col = expected_data[col_name]
            if expected_col.dtype == np.float64:
                expect_all_nan = np.all(np.isnan(expected_col))
            elif expected_col.dtype == np.int8:
                expect_all_nan = np.all(expected_col == np.iinfo("i1").min)
            elif expected_col.dtype == np.int32:
                expect_all_nan = np.all(expected_col == np.iinfo("i4").min)
            else:
                raise UnknownDataType(f"Unknown data type {expected_col.dtype}!")

            if not expect_all_nan:
                # permute expected_col if needed
                if expected_col.ndim == 1 and actual_col.ndim == actual_col_ndim:
                    if col_name == "u_angle":
                        # should be 120 and 240 degree lagging
                        expected_col = np.stack(
                            (expected_col, expected_col - 2.0 / 3.0 * np.pi, expected_col + 2.0 / 3.0 * np.pi), axis=-1
                        )
                        # reset angle to -pi to +pi range
                        expected_col = np.angle(np.exp(1j * expected_col))
                    else:
                        expected_col = expected_col.reshape(-1, 1)
                # for u_angle, the angle needs to be normalized
                # because any global angle shift is acceptable
                if col_name == "u_angle":
                    # set the u_angle of 0-th entry to zero
                    actual_col = actual_col - actual_col.ravel()[0]
                    expected_col = expected_col - expected_col.ravel()[0]

                if col_name.endswith("_angle"):
                    magnitude_name = col_name[: -len("_angle")]
                    if np.all(np.isnan(expected_data[magnitude_name])):
                        continue
                    actual_col = actual[key][magnitude_name] * np.exp(1j * actual_col)
                    expected_col = expected_data[magnitude_name] * np.exp(1j * expected_col)

                # Define the absolute tolerance based on the field name
                p = "default"
                if isinstance(atol, dict):
                    a = atol[p]
                    for pattern, tolerance in atol.items():
                        if re.fullmatch(pattern, col_name):
                            a = tolerance
                            p = f"r'{pattern}'"
                            break
                else:
                    a = atol

                assert np.allclose(actual_col, expected_col, rtol=rtol, atol=a), (
                    f"Not all values match for {key}.{col_name} (rtol={rtol}, atol={a}, pattern={p})"
                    f"\nActual:     {actual_col}"
                    f"\nExpected:   {expected_col}"
                    f"\nDifference: {actual_col - expected_col}"
                    f"\nMatches:    {np.isclose(actual_col, expected_col, rtol=rtol, atol=a)}"
                )


def convert_python_to_numpy(data: PythonDataset, data_type: DatasetType) -> Dataset:
    """
    Convert native python data to internal numpy

    Args:
        data: data in dict or list
        data_type: type of data: input, update, sym_output, or asym_output

    Returns:
        A single or batch dataset for power-grid-model
    """
    return json_deserialize(
        json.dumps(
            {
                "version": "1.0",
                "is_batch": isinstance(data, list),
                "attributes": {},
                "type": data_type,
                "data": data,
            }
        )
    )



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0


import itertools

import numpy as np
import pytest

from power_grid_model import (
    Branch3Side,
    BranchSide,
    ComponentType,
    DatasetType,
    LoadGenType,
    MeasuredTerminalType,
    WindingType,
    initialize_array,
)
from power_grid_model._core.enum import AngleMeasurementType
from power_grid_model._core.utils import compatibility_convert_row_columnar_dataset
from power_grid_model.enum import CalculationType, ComponentAttributeFilterOptions, FaultPhase, FaultType
from power_grid_model.validation import validate_input_data
from power_grid_model.validation.errors import (
    FaultPhaseError,
    InvalidAssociatedEnumValueError,
    InvalidEnumValueError,
    InvalidIdError,
    MixedPowerCurrentSensorError,
    MultiComponentNotUniqueError,
    MultiFieldValidationError,
    NotBetweenError,
    NotBetweenOrAtError,
    NotBooleanError,
    NotGreaterOrEqualError,
    NotGreaterThanError,
    NotIdenticalError,
    NotLessThanError,
    NotUniqueError,
    TwoValuesZeroError,
    UnsupportedMeasuredTerminalType,
)
from power_grid_model.validation.utils import _nan_type


@pytest.fixture
def original_data() -> dict[ComponentType, np.ndarray]:
    node = initialize_array(DatasetType.input, ComponentType.node, 4)
    node["id"] = [0, 2, 1, 2]
    node["u_rated"] = [10.5e3, 10.5e3, 0, 10.5e3]

    line = initialize_array(DatasetType.input, ComponentType.line, 3)
    line["id"] = [3, 4, 5]
    line["from_node"] = [0, -1, 2]
    line["to_node"] = [2, 1, 8]
    line["from_status"] = [0, 1, -2]
    line["to_status"] = [0, 4, 1]
    line["r1"] = [0, 100, 1]
    line["x1"] = [0, 5, 0]
    line["r0"] = [10, 0, 0]
    line["x0"] = [0, 0, 50]
    line["i_n"] = [-3, 0, 50]

    asym_line = initialize_array(DatasetType.input, ComponentType.asym_line, 4)
    asym_line["id"] = [55, 56, 57, 58]
    asym_line["from_node"] = [0, 1, 0, 0]
    asym_line["to_node"] = [1, 2, 2, 1]
    asym_line["from_status"] = [1, 1, 1, 1]
    asym_line["to_status"] = [1, 1, 1, 1]
    asym_line["r_aa"] = [-1, 2, 2, 2]
    asym_line["r_ba"] = [-1, 2, 2, 2]
    asym_line["r_bb"] = [-1, 2, 2, 2]
    asym_line["r_ca"] = [-1, 2, 2, 2]
    asym_line["r_cb"] = [-1, 2, 2, 2]
    asym_line["r_cc"] = [-1, 2, 2, 2]
    asym_line["r_na"] = [-1, 2, 2, 2]
    asym_line["r_nb"] = [-1, 2, 2, 2]
    asym_line["r_nc"] = [-1, 2, 2, 2]
    asym_line["r_nn"] = [-1, np.nan, 2, 2]
    asym_line["x_aa"] = [-1, 2, 2, 2]
    asym_line["x_ba"] = [-1, 2, 2, 2]
    asym_line["x_bb"] = [-1, 2, 2, 2]
    asym_line["x_ca"] = [-1, 2, 2, 2]
    asym_line["x_cb"] = [-1, 2, 2, 2]
    asym_line["x_cc"] = [-1, 2, 2, 2]
    asym_line["x_na"] = [-1, 2, np.nan, 2]
    asym_line["x_nb"] = [-1, 2, 2, 2]
    asym_line["x_nc"] = [-1, 2, 2, 2]
    asym_line["x_nn"] = [-1, 2, 2, 2]
    asym_line["c_aa"] = [-1, np.nan, 2, np.nan]
    asym_line["c_ba"] = [-1, np.nan, 2, np.nan]
    asym_line["c_bb"] = [-1, np.nan, 2, np.nan]
    asym_line["c_ca"] = [-1, np.nan, 2, np.nan]
    asym_line["c_cb"] = [-1, np.nan, 2, np.nan]
    asym_line["c_cc"] = [-1, np.nan, 2, 2]
    asym_line["c0"] = [-1, np.nan, np.nan, np.nan]
    asym_line["c1"] = [-1, np.nan, np.nan, np.nan]
    asym_line["i_n"] = [50, 50, 50, 50]

    generic_branch = initialize_array(DatasetType.input, ComponentType.generic_branch, 1)
    generic_branch["id"] = [6]
    generic_branch["from_node"] = [1]
    generic_branch["to_node"] = [2]
    generic_branch["from_status"] = [1]
    generic_branch["to_status"] = [1]
    generic_branch["r1"] = [0.129059]
    generic_branch["x1"] = [16.385859]
    generic_branch["g1"] = [8.692e-7]
    generic_branch["b1"] = [-2.336e-7]
    generic_branch["k"] = [0.0]
    generic_branch["theta"] = [0.0]
    generic_branch["sn"] = [-10.0]

    link = initialize_array(DatasetType.input, ComponentType.link, 2)
    link["id"] = [12, 13]
    link["from_node"] = [0, -1]
    link["to_node"] = [8, 1]
    link["from_status"] = [3, 1]
    link["to_status"] = [0, 4]

    transformer = initialize_array(DatasetType.input, ComponentType.transformer, 3)
    transformer["id"] = [1, 14, 15]
    transformer["from_node"] = [1, 7, 2]  # TODO check from node 1 to node 1
    transformer["to_node"] = [1, 8, 1]
    transformer["from_status"] = [1, 1, -1]
    transformer["to_status"] = [2, 0, 1]
    transformer["u1"] = [10500.0, 0.0, -1.0]
    transformer["u2"] = [400.0, -3.0, 0.0]
    transformer["sn"] = [630000.0, 0.0, -20.0]
    transformer["uk"] = [-1.0, 1.1, 0.9]
    transformer["pk"] = [63000.0, 0.0, -10.0]
    transformer["i0"] = [0.0368, 10.0, 0.9]
    transformer["p0"] = [63000.0, 0.0, -10.0]
    transformer["winding_from"] = [8, 0, 2]
    transformer["winding_to"] = [5, 1, 2]
    transformer["clock"] = [13, -13, 7]
    transformer["tap_side"] = [-1, 0, 1]
    transformer["tap_pos"] = [-1, 6, -4]
    transformer["tap_min"] = [-2, 4, 3]
    transformer["tap_max"] = [2, -4, -3]
    transformer["tap_nom"] = [-3, _nan_type(ComponentType.transformer, "tap_nom"), 4]
    transformer["tap_size"] = [262.5, 0.0, -10.0]
    transformer["uk_min"] = [0.0000000005, _nan_type(ComponentType.transformer, "uk_min"), 0.9]
    transformer["uk_max"] = [0.0000000005, _nan_type(ComponentType.transformer, "uk_max"), 0.8]
    transformer["pk_min"] = [300.0, 0.0, _nan_type(ComponentType.transformer, "pk_min")]
    transformer["pk_max"] = [400.0, -0.1, _nan_type(ComponentType.transformer, "pk_max")]

    three_winding_transformer = initialize_array(DatasetType.input, ComponentType.three_winding_transformer, 4)
    three_winding_transformer["id"] = [1, 28, 29, 30]
    three_winding_transformer["node_1"] = [0, 1, 9, 2]
    three_winding_transformer["node_2"] = [1, 15, 1, 0]
    three_winding_transformer["node_3"] = [1, 2, 12, 1]
    three_winding_transformer["status_1"] = [1, 5, 1, 1]
    three_winding_transformer["status_2"] = [2, 1, 1, 1]
    three_winding_transformer["status_3"] = [1, 0, -1, 0]
    three_winding_transformer["u1"] = [-100, 0, 200, 100]
    three_winding_transformer["u2"] = [0, -200, 100, 200]
    three_winding_transformer["u3"] = [-100, 0, 400, 300]
    three_winding_transformer["sn_1"] = [0, -1200, 100, 300]
    three_winding_transformer["sn_2"] = [-1000, 0, 200, 200]
    three_winding_transformer["sn_3"] = [0, -2300, 300, 100]
    three_winding_transformer["uk_12"] = [-1, 1.1, 0.05, 0.1]
    three_winding_transformer["uk_13"] = [-2, 1.2, 0.3, 0.2]
    three_winding_transformer["uk_23"] = [-1.5, 1, 0.15, 0.2]
    three_winding_transformer["pk_12"] = [-450, 100, 10, 40]
    three_winding_transformer["pk_13"] = [-40, 50, 40, 50]
    three_winding_transformer["pk_23"] = [-120, 1, 40, 30]
    three_winding_transformer["i0"] = [-0.5, 1.8, 0.3, 0.6]
    three_winding_transformer["p0"] = [-100, 410, 60, 40]
    three_winding_transformer["winding_1"] = [15, -1, 0, 2]
    three_winding_transformer["winding_2"] = [19, -2, 1, 3]
    three_winding_transformer["winding_3"] = [-2, 13, 2, 2]
    three_winding_transformer["clock_12"] = [-12, 24, 4, 3]
    three_winding_transformer["clock_13"] = [-30, 40, 3, 4]
    three_winding_transformer["tap_side"] = [-1, 9, 1, 0]
    three_winding_transformer["tap_pos"] = [50, -24, 5, 3]
    three_winding_transformer["tap_min"] = [-10, -10, -10, -10]
    three_winding_transformer["tap_max"] = [10, 10, 10, 10]
    three_winding_transformer["tap_size"] = [-12, 0, 3, 130]
    three_winding_transformer["tap_nom"] = [-12, 41, _nan_type(ComponentType.three_winding_transformer, "tap_nom"), 0]
    three_winding_transformer["uk_12_min"] = [
        _nan_type(ComponentType.three_winding_transformer, "uk_12_min"),
        1.1,
        0.05,
        _nan_type(ComponentType.three_winding_transformer, "uk_12_min"),
    ]
    three_winding_transformer["uk_13_min"] = [
        _nan_type(ComponentType.three_winding_transformer, "uk_13_min"),
        1.2,
        0.3,
        _nan_type(ComponentType.three_winding_transformer, "uk_13_min"),
    ]
    three_winding_transformer["uk_23_min"] = [
        _nan_type(ComponentType.three_winding_transformer, "uk_23_min"),
        1,
        0.15,
        _nan_type(ComponentType.three_winding_transformer, "uk_23_min"),
    ]
    three_winding_transformer["pk_12_min"] = [
        -450,
        _nan_type(ComponentType.three_winding_transformer, "pk_12_min"),
        10,
        40,
    ]
    three_winding_transformer["pk_13_min"] = [
        -40,
        _nan_type(ComponentType.three_winding_transformer, "pk_13_min"),
        40,
        50,
    ]
    three_winding_transformer["pk_23_min"] = [
        -120,
        _nan_type(ComponentType.three_winding_transformer, "pk_23_min"),
        40,
        30,
    ]
    three_winding_transformer["uk_12_max"] = [
        _nan_type(ComponentType.three_winding_transformer, "uk_12_max"),
        1.1,
        0.05,
        _nan_type(ComponentType.three_winding_transformer, "uk_12_max"),
    ]
    three_winding_transformer["uk_13_max"] = [
        _nan_type(ComponentType.three_winding_transformer, "uk_13_max"),
        1.2,
        0.3,
        _nan_type(ComponentType.three_winding_transformer, "uk_13_max"),
    ]
    three_winding_transformer["uk_23_max"] = [
        _nan_type(ComponentType.three_winding_transformer, "uk_23_max"),
        1,
        0.15,
        _nan_type(ComponentType.three_winding_transformer, "uk_23_max"),
    ]
    three_winding_transformer["pk_12_max"] = [
        -450,
        _nan_type(ComponentType.three_winding_transformer, "pk_12_max"),
        10,
        40,
    ]
    three_winding_transformer["pk_13_max"] = [
        -40,
        _nan_type(ComponentType.three_winding_transformer, "pk_12_max"),
        40,
        50,
    ]
    three_winding_transformer["pk_23_max"] = [
        -120,
        _nan_type(ComponentType.three_winding_transformer, "pk_12_max"),
        40,
        30,
    ]

    transformer_tap_regulator = initialize_array(DatasetType.input, ComponentType.transformer_tap_regulator, 5)
    transformer_tap_regulator["id"] = [51, 52, 53, 54, 1]
    transformer_tap_regulator["status"] = [0, -1, 2, 1, 5]
    transformer_tap_regulator["regulated_object"] = [14, 15, 28, 14, 2]
    transformer_tap_regulator["control_side"] = [1, 2, 0, 0, 60]
    transformer_tap_regulator["u_set"] = [100, -100, 100, 100, 100]
    transformer_tap_regulator["u_band"] = [100, -4, 100, 100, 0]
    transformer_tap_regulator["line_drop_compensation_r"] = [0.0, -1.0, 1.0, 0.0, 2.0]
    transformer_tap_regulator["line_drop_compensation_x"] = [0.0, 4.0, 2.0, 0.0, -4.0]

    source = initialize_array(DatasetType.input, ComponentType.source, 3)
    source["id"] = [16, 17, 1]
    source["node"] = [10, 1, 2]
    source["status"] = [0, -1, 2]
    source["u_ref"] = [-10.0, 0.0, 100.0]
    source["sk"] = [0.0, 100.0, -20.0]
    source["rx_ratio"] = [0.0, -30.0, 300.0]
    source["z01_ratio"] = [-1.0, 0.0, 200.0]

    shunt = initialize_array(DatasetType.input, ComponentType.shunt, 3)
    shunt["id"] = [18, 19, 1]
    shunt["node"] = [10, 1, 2]
    shunt["status"] = [0, -1, 2]

    sym_load = initialize_array(DatasetType.input, ComponentType.sym_load, 3)
    sym_load["id"] = [1, 20, 21]
    sym_load["type"] = [1, 0, 5]
    sym_load["node"] = [10, 1, 2]
    sym_load["status"] = [0, -1, 2]

    sym_gen = initialize_array(DatasetType.input, ComponentType.sym_gen, 3)
    sym_gen["id"] = [1, 22, 23]
    sym_gen["type"] = [2, -1, 1]
    sym_gen["node"] = [10, 1, 2]
    sym_gen["status"] = [0, -1, 2]

    asym_load = initialize_array(DatasetType.input, ComponentType.asym_load, 3)
    asym_load["id"] = [1, 24, 25]
    asym_load["type"] = [5, 0, 2]
    asym_load["node"] = [10, 1, 2]
    asym_load["status"] = [0, -1, 2]

    asym_gen = initialize_array(DatasetType.input, ComponentType.asym_gen, 3)
    asym_gen["id"] = [1, 26, 27]
    asym_gen["type"] = [-1, 5, 2]
    asym_gen["node"] = [10, 1, 2]
    asym_gen["status"] = [0, -1, 2]

    sym_voltage_sensor = initialize_array(DatasetType.input, ComponentType.sym_voltage_sensor, 5)
    sym_voltage_sensor["id"] = range(107, 107 + 5)
    sym_voltage_sensor["measured_object"] = [2, 3, 1, 200, 2]
    sym_voltage_sensor["u_measured"] = [0.0, 10.4e3, 10.6e3, -20.0, 1e4]
    sym_voltage_sensor["u_sigma"] = [1.0, np.nan, 0.0, -1.0, np.inf]

    asym_voltage_sensor = initialize_array(DatasetType.input, ComponentType.asym_voltage_sensor, 5)
    asym_voltage_sensor["id"] = range(107, 107 + 5)
    asym_voltage_sensor["measured_object"] = [2, 3, 1, 200, 2]
    asym_voltage_sensor["u_measured"] = np.array(
        [
            [10.5e3, 10.4e3, 10.6e3],
            [np.nan, np.nan, np.nan],
            [0, 0, 0],
            [-1e4, 1e4, 1e4],
            [1.0e4, 1.0e4, 1.0e4],
        ]
    )
    asym_voltage_sensor["u_sigma"] = [1.0, np.nan, 0.0, -1.0, np.inf]

    sym_power_sensor = initialize_array(DatasetType.input, ComponentType.sym_power_sensor, 15)
    sym_power_sensor["id"] = range(107, 107 + 15)
    sym_power_sensor["measured_object"] = [12, 3, 13, 200, 3] + [3] * 10
    sym_power_sensor["power_sigma"] = [1.0, np.nan, 0.0, -1.0, np.inf] * 2 + [np.nan] * 5
    sym_power_sensor["measured_terminal_type"] = [1, 1, 10, 1, 1] + [1] * 10
    sym_power_sensor["p_sigma"] = [np.nan] * 5 + [1.0, 1.0, 0.0, -1.0, np.inf] * 2
    sym_power_sensor["q_sigma"] = [np.nan] * 5 + [1.0, np.inf, 0.0, -1.0, np.inf] * 2

    asym_power_sensor = initialize_array(DatasetType.input, ComponentType.asym_power_sensor, 15)
    asym_power_sensor["id"] = range(107, 107 + 15)
    asym_power_sensor["measured_object"] = [12, 3, 13, 200, 3] + [3] * 10
    asym_power_sensor["power_sigma"] = [1.0, np.nan, 0.0, -1.0, np.inf] * 2 + [np.nan] * 5
    asym_power_sensor["measured_terminal_type"] = [1, 1, 10, 1, 1] + [1] * 10
    asym_power_sensor["p_sigma"] = [[np.nan] * 3] * 5 + [
        [1.0, 1.0, 1.0],
        [np.inf, np.inf, np.inf],
        [1.0, np.nan, 1.0],
        [1.0, 1.0, -1.0],
        [np.inf, 1.0, 1.0],
    ] * 2
    asym_power_sensor["q_sigma"] = [[np.nan] * 3] * 5 + [
        [1.0, 1.0, 1.0],
        [np.inf, np.inf, np.inf],
        [np.inf, np.nan, 1.0],
        [1.0, 1.0, -1.0],
        [np.inf, 1.0, 1.0],
    ] * 2

    sym_current_sensor = initialize_array(DatasetType.input, ComponentType.sym_current_sensor, 5)
    sym_current_sensor["id"] = range(107, 107 + 5)
    sym_current_sensor["measured_object"] = [12, 3, 13, 200, 3]
    sym_current_sensor["i_sigma"] = [1.0, np.nan, 0.0, -1.0, np.inf]
    sym_current_sensor["i_angle_sigma"] = [1.0, np.nan, 0.0, -1.0, np.inf]
    sym_current_sensor["measured_terminal_type"] = [1, 1, 10, 2, 1]
    sym_current_sensor["angle_measurement_type"] = [0, 1, 10, 2, 0]

    asym_current_sensor = initialize_array(DatasetType.input, ComponentType.asym_current_sensor, 5)
    asym_current_sensor["id"] = range(107, 107 + 5)
    asym_current_sensor["measured_object"] = [12, 3, 13, 200, 3]
    asym_current_sensor["i_sigma"] = [1.0, np.nan, 0.0, -1.0, np.inf]
    asym_current_sensor["i_angle_sigma"] = [1.0, np.nan, 0.0, -1.0, np.inf]
    asym_current_sensor["measured_terminal_type"] = [1, 1, 10, 2, 1]
    asym_current_sensor["angle_measurement_type"] = [0, 1, 10, 2, 0]

    fault = initialize_array(DatasetType.input, ComponentType.fault, 20)
    fault["id"] = [1, *list(range(32, 51))]
    fault["status"] = [0, -1, 2] + 17 * [1]
    fault["fault_type"] = 6 * [0] + 4 * [1] + 4 * [2] + 4 * [3] + [_nan_type(ComponentType.fault, "fault_type"), 4]
    fault["fault_phase"] = (
        list(range(1, 7)) + [0, 4, 5, 6] + 2 * list(range(4)) + [_nan_type(ComponentType.fault, "fault_phase"), 7]
    )
    fault["fault_object"] = [200, 3] + list(range(10, 28, 2)) + 9 * [0]
    fault["r_f"] = [-1.0, 0.0, 1.0] + 17 * [_nan_type(ComponentType.fault, "r_f")]
    fault["x_f"] = [-1.0, 0.0, 1.0] + 17 * [_nan_type(ComponentType.fault, "x_f")]
    data = {
        ComponentType.node: node,
        ComponentType.line: line,
        ComponentType.asym_line: asym_line,
        ComponentType.generic_branch: generic_branch,
        ComponentType.link: link,
        ComponentType.transformer: transformer,
        ComponentType.three_winding_transformer: three_winding_transformer,
        ComponentType.transformer_tap_regulator: transformer_tap_regulator,
        ComponentType.source: source,
        ComponentType.shunt: shunt,
        ComponentType.sym_load: sym_load,
        ComponentType.sym_gen: sym_gen,
        ComponentType.asym_load: asym_load,
        ComponentType.asym_gen: asym_gen,
        ComponentType.sym_voltage_sensor: sym_voltage_sensor,
        ComponentType.asym_voltage_sensor: asym_voltage_sensor,
        ComponentType.sym_power_sensor: sym_power_sensor,
        ComponentType.asym_power_sensor: asym_power_sensor,
        ComponentType.sym_current_sensor: sym_current_sensor,
        ComponentType.asym_current_sensor: asym_current_sensor,
        ComponentType.fault: fault,
    }
    return data  # noqa: RET504


@pytest.fixture
def original_data_columnar_all(original_data):
    return compatibility_convert_row_columnar_dataset(
        original_data, ComponentAttributeFilterOptions.everything, DatasetType.input
    )


@pytest.fixture
def original_data_columnar_relevant(original_data):
    return compatibility_convert_row_columnar_dataset(
        original_data, ComponentAttributeFilterOptions.relevant, DatasetType.input
    )


@pytest.fixture(params=["original_data", "original_data_columnar_all", "original_data_columnar_relevant"])
def input_data(request):
    return request.getfixturevalue(request.param)


def test_validate_input_data_sym_calculation(input_data):
    validation_errors = validate_input_data(input_data, symmetric=True)

    assert validation_errors is not None
    assert (
        MultiComponentNotUniqueError(
            [
                (ComponentType.asym_gen, "id"),
                (ComponentType.asym_load, "id"),
                (ComponentType.asym_current_sensor, "id"),
                (ComponentType.asym_power_sensor, "id"),
                (ComponentType.asym_voltage_sensor, "id"),
                (ComponentType.node, "id"),
                (ComponentType.shunt, "id"),
                (ComponentType.source, "id"),
                (ComponentType.sym_gen, "id"),
                (ComponentType.sym_load, "id"),
                (ComponentType.sym_current_sensor, "id"),
                (ComponentType.sym_power_sensor, "id"),
                (ComponentType.sym_voltage_sensor, "id"),
                (ComponentType.transformer, "id"),
                (ComponentType.three_winding_transformer, "id"),
                (ComponentType.fault, "id"),
                (ComponentType.transformer_tap_regulator, "id"),
            ],
            [
                (ComponentType.asym_gen, 1),
                (ComponentType.asym_load, 1),
                *[(ComponentType.asym_current_sensor, i) for i in range(107, 107 + 5)],
                *[(ComponentType.asym_power_sensor, i) for i in range(107, 107 + 15)],
                *[(ComponentType.asym_voltage_sensor, i) for i in range(107, 107 + 5)],
                (ComponentType.node, 1),
                (ComponentType.shunt, 1),
                (ComponentType.source, 1),
                (ComponentType.sym_gen, 1),
                (ComponentType.sym_load, 1),
                *[(ComponentType.sym_current_sensor, i) for i in range(107, 107 + 5)],
                *[(ComponentType.sym_power_sensor, i) for i in range(107, 107 + 15)],
                *[(ComponentType.sym_voltage_sensor, i) for i in range(107, 107 + 5)],
                (ComponentType.transformer, 1),
                (ComponentType.three_winding_transformer, 1),
                (ComponentType.fault, 1),
                (ComponentType.transformer_tap_regulator, 1),
            ],
        )
        in validation_errors
    )

    assert NotGreaterThanError(ComponentType.node, "u_rated", [1], 0) in validation_errors
    assert NotUniqueError(ComponentType.node, "id", [2, 2]) in validation_errors

    assert NotBooleanError(ComponentType.line, "from_status", [5]) in validation_errors
    assert NotBooleanError(ComponentType.line, "to_status", [4]) in validation_errors
    assert InvalidIdError(ComponentType.line, "from_node", [4], ComponentType.node) in validation_errors
    assert InvalidIdError(ComponentType.line, "to_node", [5], ComponentType.node) in validation_errors
    assert TwoValuesZeroError(ComponentType.line, ["r1", "x1"], [3]) in validation_errors
    assert TwoValuesZeroError(ComponentType.line, ["r0", "x0"], [4]) in validation_errors
    assert NotGreaterThanError(ComponentType.line, "i_n", [3, 4], 0) in validation_errors

    assert NotBooleanError(ComponentType.link, "from_status", [12]) in validation_errors
    assert NotBooleanError(ComponentType.link, "to_status", [13]) in validation_errors
    assert InvalidIdError(ComponentType.link, "from_node", [13], ComponentType.node) in validation_errors
    assert InvalidIdError(ComponentType.link, "to_node", [12], ComponentType.node) in validation_errors

    assert NotBooleanError(ComponentType.transformer, "from_status", [15]) in validation_errors
    assert NotBooleanError(ComponentType.transformer, "to_status", [1]) in validation_errors
    assert InvalidIdError(ComponentType.transformer, "from_node", [14], ComponentType.node) in validation_errors
    assert InvalidIdError(ComponentType.transformer, "to_node", [14], ComponentType.node) in validation_errors
    assert NotGreaterThanError(ComponentType.transformer, "u1", [14, 15], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.transformer, "u2", [14, 15], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.transformer, "sn", [14, 15], 0) in validation_errors
    assert NotGreaterOrEqualError(ComponentType.transformer, "uk", [1], "pk/sn") in validation_errors
    assert NotBetweenError(ComponentType.transformer, "uk", [1, 14], (0, 1)) in validation_errors
    assert NotGreaterOrEqualError(ComponentType.transformer, "pk", [15], 0) in validation_errors
    assert NotGreaterOrEqualError(ComponentType.transformer, "i0", [1], "p0/sn") in validation_errors
    assert NotLessThanError(ComponentType.transformer, "i0", [14], 1) in validation_errors
    assert NotGreaterOrEqualError(ComponentType.transformer, "p0", [15], 0) in validation_errors
    assert NotBetweenOrAtError(ComponentType.transformer, "clock", [1, 14], (-12, 12)) in validation_errors
    assert (
        NotBetweenOrAtError(ComponentType.transformer, "tap_pos", [14, 15], ("tap_min", "tap_max")) in validation_errors
    )
    assert (
        NotBetweenOrAtError(ComponentType.transformer, "tap_nom", [1, 15], ("tap_min", "tap_max")) in validation_errors
    )
    assert NotGreaterOrEqualError(ComponentType.transformer, "tap_size", [15], 0) in validation_errors
    assert NotGreaterOrEqualError(ComponentType.transformer, "uk_min", [1], "pk_min/sn") in validation_errors
    assert NotBetweenError(ComponentType.transformer, "uk_min", [14], (0, 1)) in validation_errors
    assert NotGreaterOrEqualError(ComponentType.transformer, "uk_max", [1], "pk_max/sn") in validation_errors
    assert NotBetweenError(ComponentType.transformer, "uk_max", [14], (0, 1)) in validation_errors
    assert NotGreaterOrEqualError(ComponentType.transformer, "pk_min", [15], 0) in validation_errors
    assert NotGreaterOrEqualError(ComponentType.transformer, "pk_max", [14, 15], 0) in validation_errors
    assert InvalidEnumValueError(ComponentType.transformer, "winding_from", [1], WindingType) in validation_errors
    assert InvalidEnumValueError(ComponentType.transformer, "winding_to", [1], WindingType) in validation_errors
    assert InvalidEnumValueError(ComponentType.transformer, "tap_side", [1], BranchSide) in validation_errors

    assert InvalidIdError(ComponentType.source, "node", [16], ComponentType.node) in validation_errors
    assert NotBooleanError(ComponentType.source, "status", [17, 1]) in validation_errors
    assert NotGreaterThanError(ComponentType.source, "u_ref", [16, 17], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.source, "sk", [16, 1], 0) in validation_errors
    assert NotGreaterOrEqualError(ComponentType.source, "rx_ratio", [17], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.source, "z01_ratio", [16, 17], 0) in validation_errors

    assert InvalidIdError(ComponentType.shunt, "node", [18], ComponentType.node) in validation_errors
    assert NotBooleanError(ComponentType.shunt, "status", [19, 1]) in validation_errors

    assert InvalidIdError(ComponentType.sym_load, "node", [1], ComponentType.node) in validation_errors
    assert NotBooleanError(ComponentType.sym_load, "status", [20, 21]) in validation_errors
    assert InvalidEnumValueError(ComponentType.sym_load, "type", [21], LoadGenType) in validation_errors

    assert InvalidIdError(ComponentType.sym_gen, "node", [1], ComponentType.node) in validation_errors
    assert NotBooleanError(ComponentType.sym_gen, "status", [22, 23]) in validation_errors
    assert InvalidEnumValueError(ComponentType.sym_gen, "type", [22], LoadGenType) in validation_errors

    assert InvalidIdError(ComponentType.asym_load, "node", [1], ComponentType.node) in validation_errors
    assert NotBooleanError(ComponentType.asym_load, "status", [24, 25]) in validation_errors
    assert InvalidEnumValueError(ComponentType.asym_load, "type", [1], LoadGenType) in validation_errors

    assert InvalidIdError(ComponentType.asym_gen, "node", [1], ComponentType.node) in validation_errors
    assert NotBooleanError(ComponentType.asym_gen, "status", [26, 27]) in validation_errors
    assert InvalidEnumValueError(ComponentType.asym_gen, "type", [1, 26], LoadGenType) in validation_errors

    assert InvalidIdError(ComponentType.sym_voltage_sensor, "measured_object", [108, 110], "node") in validation_errors
    assert NotGreaterThanError(ComponentType.sym_voltage_sensor, "u_measured", [107, 110], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.sym_voltage_sensor, "u_sigma", [109, 110], 0) in validation_errors
    # TODO check if u_sigma = np.nan is detected with missing values

    assert InvalidIdError(ComponentType.asym_voltage_sensor, "measured_object", [108, 110], "node") in validation_errors
    assert NotGreaterThanError(ComponentType.asym_voltage_sensor, "u_measured", [109, 110], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.asym_voltage_sensor, "u_sigma", [109, 110], 0) in validation_errors

    assert (
        InvalidIdError(
            ComponentType.sym_power_sensor,
            "measured_object",
            [107, 109, 110],
            [
                ComponentType.node,
                ComponentType.line,
                ComponentType.asym_line,
                ComponentType.generic_branch,
                ComponentType.transformer,
                ComponentType.three_winding_transformer,
                ComponentType.source,
                ComponentType.shunt,
                ComponentType.sym_load,
                ComponentType.asym_load,
                ComponentType.sym_gen,
                ComponentType.asym_gen,
            ],
        )
        in validation_errors
    )
    assert (
        NotGreaterThanError(ComponentType.sym_power_sensor, "power_sigma", [109, 110, 114, 115], 0) in validation_errors
    )
    assert (
        InvalidIdError(
            ComponentType.sym_power_sensor,
            "measured_object",
            [107, 110],
            [ComponentType.line, ComponentType.asym_line, ComponentType.generic_branch, ComponentType.transformer],
            {"measured_terminal_type": MeasuredTerminalType.branch_to},
        )
        in validation_errors
    )
    assert (
        InvalidEnumValueError(ComponentType.sym_power_sensor, "measured_terminal_type", [109], MeasuredTerminalType)
        in validation_errors
    )

    assert (
        InvalidIdError(
            ComponentType.asym_power_sensor,
            "measured_object",
            [107, 109, 110],
            [
                ComponentType.node,
                ComponentType.line,
                ComponentType.asym_line,
                ComponentType.generic_branch,
                ComponentType.transformer,
                ComponentType.three_winding_transformer,
                ComponentType.source,
                ComponentType.shunt,
                ComponentType.sym_load,
                ComponentType.asym_load,
                ComponentType.sym_gen,
                ComponentType.asym_gen,
            ],
        )
        in validation_errors
    )
    assert (
        NotGreaterThanError(ComponentType.asym_power_sensor, "power_sigma", [109, 110, 114, 115], 0)
        in validation_errors
    )
    assert (
        InvalidIdError(
            ComponentType.asym_power_sensor,
            "measured_object",
            [107, 110],
            [ComponentType.line, ComponentType.asym_line, ComponentType.generic_branch, ComponentType.transformer],
            {"measured_terminal_type": MeasuredTerminalType.branch_to},
        )
        in validation_errors
    )
    assert (
        InvalidEnumValueError(ComponentType.asym_power_sensor, "measured_terminal_type", [109], MeasuredTerminalType)
        in validation_errors
    )

    assert (
        InvalidIdError(
            ComponentType.sym_current_sensor,
            "measured_object",
            [107, 109, 110],
            [
                ComponentType.line,
                ComponentType.asym_line,
                ComponentType.generic_branch,
                ComponentType.transformer,
                ComponentType.three_winding_transformer,
            ],
        )
        in validation_errors
    )
    assert NotGreaterThanError(ComponentType.sym_current_sensor, "i_sigma", [109, 110], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.sym_current_sensor, "i_angle_sigma", [109, 110], 0) in validation_errors
    assert (
        InvalidIdError(
            ComponentType.sym_current_sensor,
            "measured_object",
            [107],
            [ComponentType.line, ComponentType.asym_line, ComponentType.generic_branch, ComponentType.transformer],
            {"measured_terminal_type": MeasuredTerminalType.branch_to},
        )
        in validation_errors
    )
    assert (
        InvalidEnumValueError(ComponentType.sym_current_sensor, "measured_terminal_type", [109], MeasuredTerminalType)
        in validation_errors
    )
    assert (
        UnsupportedMeasuredTerminalType(
            ComponentType.sym_current_sensor,
            "measured_terminal_type",
            [109, 110],
            [
                MeasuredTerminalType.branch_from,
                MeasuredTerminalType.branch_to,
                MeasuredTerminalType.branch3_1,
                MeasuredTerminalType.branch3_2,
                MeasuredTerminalType.branch3_3,
            ],
        )
        in validation_errors
    )
    assert (
        InvalidEnumValueError(
            ComponentType.sym_current_sensor, "angle_measurement_type", [109, 110], AngleMeasurementType
        )
        in validation_errors
    )

    assert (
        InvalidIdError(
            ComponentType.asym_current_sensor,
            "measured_object",
            [107, 109, 110],
            [
                ComponentType.line,
                ComponentType.asym_line,
                ComponentType.generic_branch,
                ComponentType.transformer,
                ComponentType.three_winding_transformer,
            ],
        )
        in validation_errors
    )
    assert NotGreaterThanError(ComponentType.asym_current_sensor, "i_sigma", [109, 110], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.asym_current_sensor, "i_angle_sigma", [109, 110], 0) in validation_errors
    assert (
        InvalidIdError(
            ComponentType.asym_current_sensor,
            "measured_object",
            [107],
            [ComponentType.line, ComponentType.asym_line, ComponentType.generic_branch, ComponentType.transformer],
            {"measured_terminal_type": MeasuredTerminalType.branch_to},
        )
        in validation_errors
    )
    assert (
        InvalidEnumValueError(ComponentType.asym_current_sensor, "measured_terminal_type", [109], MeasuredTerminalType)
        in validation_errors
    )
    assert (
        UnsupportedMeasuredTerminalType(
            ComponentType.asym_current_sensor,
            "measured_terminal_type",
            [109, 110],
            [
                MeasuredTerminalType.branch_from,
                MeasuredTerminalType.branch_to,
                MeasuredTerminalType.branch3_1,
                MeasuredTerminalType.branch3_2,
                MeasuredTerminalType.branch3_3,
            ],
        )
        in validation_errors
    )
    assert (
        InvalidEnumValueError(
            ComponentType.asym_current_sensor, "angle_measurement_type", [109, 110], AngleMeasurementType
        )
        in validation_errors
    )
    for power_sensor_type, current_sensor_type in itertools.product(
        [ComponentType.sym_power_sensor, ComponentType.asym_power_sensor],
        [ComponentType.sym_current_sensor, ComponentType.asym_current_sensor],
    ):
        assert (
            MixedPowerCurrentSensorError(
                [
                    (power_sensor_type, "measured_object"),
                    (power_sensor_type, "measured_terminal_type"),
                    (current_sensor_type, "measured_object"),
                    (current_sensor_type, "measured_terminal_type"),
                ],
                [
                    *[(power_sensor_type, i) for i in range(107, 107 + 3)],
                    # 110 is skipped because it points to an invalid measured terminal type
                    *[(power_sensor_type, i) for i in range(111, 111 + 11)],
                    *[(current_sensor_type, i) for i in range(107, 107 + 3)],
                    # 110 is skipped because it points to an invalid measured terminal type
                    (current_sensor_type, 111),
                ],
            )
            in validation_errors
        )

    assert NotGreaterOrEqualError(ComponentType.transformer, "uk_max", [15], "uk_min") not in validation_errors

    assert NotBooleanError(ComponentType.fault, "status", [32, 33]) in validation_errors
    assert (
        InvalidIdError(ComponentType.fault, "fault_object", [1, *list(range(32, 42))], [ComponentType.node])
        in validation_errors
    )


def test_validate_three_winding_transformer(input_data):
    validation_errors = validate_input_data(input_data, symmetric=True)
    assert validation_errors is not None
    assert NotBooleanError(ComponentType.three_winding_transformer, "status_1", [28]) in validation_errors
    assert NotBooleanError(ComponentType.three_winding_transformer, "status_2", [1]) in validation_errors
    assert NotBooleanError(ComponentType.three_winding_transformer, "status_3", [29]) in validation_errors
    assert (
        InvalidIdError(ComponentType.three_winding_transformer, "node_1", [29], ComponentType.node) in validation_errors
    )
    assert (
        InvalidIdError(ComponentType.three_winding_transformer, "node_2", [28], ComponentType.node) in validation_errors
    )
    assert (
        InvalidIdError(ComponentType.three_winding_transformer, "node_3", [29], ComponentType.node) in validation_errors
    )
    assert NotGreaterThanError(ComponentType.three_winding_transformer, "u1", [1, 28], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.three_winding_transformer, "u2", [1, 28], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.three_winding_transformer, "u3", [1, 28], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.three_winding_transformer, "sn_1", [1, 28], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.three_winding_transformer, "sn_2", [1, 28], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.three_winding_transformer, "sn_3", [1, 28], 0) in validation_errors
    assert (
        NotGreaterOrEqualError(ComponentType.three_winding_transformer, "uk_12", [29, 30], "pk_12/sn_1")
        in validation_errors
    )
    assert (
        NotGreaterOrEqualError(ComponentType.three_winding_transformer, "uk_12", [1, 30], "pk_12/sn_2")
        in validation_errors
    )
    assert (
        NotGreaterOrEqualError(ComponentType.three_winding_transformer, "uk_13", [29], "pk_13/sn_1")
        in validation_errors
    )
    assert (
        NotGreaterOrEqualError(ComponentType.three_winding_transformer, "uk_13", [30], "pk_13/sn_3")
        in validation_errors
    )
    assert (
        NotGreaterOrEqualError(ComponentType.three_winding_transformer, "uk_23", [1, 29], "pk_23/sn_2")
        in validation_errors
    )
    assert (
        NotGreaterOrEqualError(ComponentType.three_winding_transformer, "uk_23", [30], "pk_23/sn_3")
        in validation_errors
    )
    assert NotBetweenError(ComponentType.three_winding_transformer, "uk_12", [1, 28], (0, 1)) in validation_errors
    assert NotBetweenError(ComponentType.three_winding_transformer, "uk_13", [1, 28], (0, 1)) in validation_errors
    assert NotBetweenError(ComponentType.three_winding_transformer, "uk_23", [1, 28], (0, 1)) in validation_errors
    assert NotGreaterOrEqualError(ComponentType.three_winding_transformer, "pk_12", [1], 0) in validation_errors
    assert NotGreaterOrEqualError(ComponentType.three_winding_transformer, "pk_13", [1], 0) in validation_errors
    assert NotGreaterOrEqualError(ComponentType.three_winding_transformer, "pk_23", [1], 0) in validation_errors
    assert NotGreaterOrEqualError(ComponentType.three_winding_transformer, "i0", [29], "p0/sn_1") in validation_errors
    assert NotLessThanError(ComponentType.three_winding_transformer, "i0", [28], 1) in validation_errors
    assert NotGreaterOrEqualError(ComponentType.three_winding_transformer, "p0", [1], 0) in validation_errors
    assert (
        NotBetweenOrAtError(ComponentType.three_winding_transformer, "clock_13", [1, 28], (-12, 12))
        in validation_errors
    )
    assert (
        NotBetweenOrAtError(ComponentType.three_winding_transformer, "tap_pos", [1, 28], ("tap_min", "tap_max"))
        in validation_errors
    )
    assert (
        NotBetweenOrAtError(ComponentType.three_winding_transformer, "tap_nom", [1, 28], ("tap_min", "tap_max"))
        in validation_errors
    )
    assert NotGreaterOrEqualError(ComponentType.three_winding_transformer, "tap_size", [1], 0) in validation_errors
    assert (
        InvalidEnumValueError(ComponentType.three_winding_transformer, "winding_1", [1, 28], WindingType)
        in validation_errors
    )
    assert (
        InvalidEnumValueError(ComponentType.three_winding_transformer, "winding_2", [1, 28], WindingType)
        in validation_errors
    )
    assert (
        InvalidEnumValueError(ComponentType.three_winding_transformer, "winding_3", [1, 28], WindingType)
        in validation_errors
    )
    assert (
        InvalidEnumValueError(ComponentType.three_winding_transformer, "tap_side", [1, 28], Branch3Side)
        in validation_errors
    )


def test_validate_three_winding_transformer_ukpkminmax(input_data):
    validation_errors = validate_input_data(input_data, symmetric=False)
    assert validation_errors is not None
    assert (
        NotGreaterOrEqualError(ComponentType.three_winding_transformer, "uk_12_min", [29, 30], "pk_12_min/sn_1")
        in validation_errors
    )
    assert (
        NotGreaterOrEqualError(ComponentType.three_winding_transformer, "uk_12_min", [1, 30], "pk_12_min/sn_2")
        in validation_errors
    )
    assert (
        NotGreaterOrEqualError(ComponentType.three_winding_transformer, "uk_13_min", [29], "pk_13_min/sn_1")
        in validation_errors
    )
    assert (
        NotGreaterOrEqualError(ComponentType.three_winding_transformer, "uk_13_min", [30], "pk_13_min/sn_3")
        in validation_errors
    )
    assert (
        NotGreaterOrEqualError(ComponentType.three_winding_transformer, "uk_23_min", [1, 29], "pk_23_min/sn_2")
        in validation_errors
    )
    assert (
        NotGreaterOrEqualError(ComponentType.three_winding_transformer, "uk_23_min", [30], "pk_23_min/sn_3")
        in validation_errors
    )
    assert NotBetweenError(ComponentType.three_winding_transformer, "uk_12_min", [1, 28], (0, 1)) in validation_errors
    assert NotBetweenError(ComponentType.three_winding_transformer, "uk_13_min", [1, 28], (0, 1)) in validation_errors
    assert NotBetweenError(ComponentType.three_winding_transformer, "uk_23_min", [1, 28], (0, 1)) in validation_errors
    assert (
        NotGreaterOrEqualError(ComponentType.three_winding_transformer, "uk_12_max", [29, 30], "pk_12_max/sn_1")
        in validation_errors
    )
    assert (
        NotGreaterOrEqualError(ComponentType.three_winding_transformer, "uk_12_max", [1, 30], "pk_12_max/sn_2")
        in validation_errors
    )
    assert (
        NotGreaterOrEqualError(ComponentType.three_winding_transformer, "uk_13_max", [29], "pk_13_max/sn_1")
        in validation_errors
    )
    assert (
        NotGreaterOrEqualError(ComponentType.three_winding_transformer, "uk_13_max", [30], "pk_13_max/sn_3")
        in validation_errors
    )
    assert (
        NotGreaterOrEqualError(ComponentType.three_winding_transformer, "uk_23_max", [1, 29], "pk_23_max/sn_2")
        in validation_errors
    )
    assert (
        NotGreaterOrEqualError(ComponentType.three_winding_transformer, "uk_23_max", [30], "pk_23_max/sn_3")
        in validation_errors
    )
    assert NotBetweenError(ComponentType.three_winding_transformer, "uk_12_max", [1, 28], (0, 1)) in validation_errors
    assert NotBetweenError(ComponentType.three_winding_transformer, "uk_13_max", [1, 28], (0, 1)) in validation_errors
    assert NotBetweenError(ComponentType.three_winding_transformer, "uk_23_max", [1, 28], (0, 1)) in validation_errors
    assert NotGreaterOrEqualError(ComponentType.three_winding_transformer, "pk_12_min", [1], 0) in validation_errors
    assert NotGreaterOrEqualError(ComponentType.three_winding_transformer, "pk_13_min", [1], 0) in validation_errors
    assert NotGreaterOrEqualError(ComponentType.three_winding_transformer, "pk_23_min", [1], 0) in validation_errors
    assert NotGreaterOrEqualError(ComponentType.three_winding_transformer, "pk_12_max", [1], 0) in validation_errors
    assert NotGreaterOrEqualError(ComponentType.three_winding_transformer, "pk_13_max", [1], 0) in validation_errors
    assert NotGreaterOrEqualError(ComponentType.three_winding_transformer, "pk_23_max", [1], 0) in validation_errors


def test_validate_input_data_transformer_tap_regulator(input_data):
    validation_errors = validate_input_data(input_data, calculation_type=CalculationType.power_flow)
    assert validation_errors is not None
    assert NotBooleanError(ComponentType.transformer_tap_regulator, "status", [52, 1, 53]) in validation_errors
    assert (
        InvalidIdError(
            ComponentType.transformer_tap_regulator,
            "regulated_object",
            [1],
            [ComponentType.transformer, ComponentType.three_winding_transformer],
        )
        in validation_errors
    )
    assert (
        InvalidEnumValueError(ComponentType.transformer_tap_regulator, "control_side", [1], [BranchSide, Branch3Side])
        in validation_errors
    )
    assert (
        InvalidAssociatedEnumValueError(
            ComponentType.transformer_tap_regulator, ["control_side", "regulated_object"], [52], [BranchSide]
        )
        in validation_errors
    )
    assert NotGreaterOrEqualError(ComponentType.transformer_tap_regulator, "u_set", [52], 0.0) in validation_errors
    assert NotGreaterThanError(ComponentType.transformer_tap_regulator, "u_band", [52, 1], 0.0) in validation_errors
    assert (
        NotGreaterOrEqualError(ComponentType.transformer_tap_regulator, "line_drop_compensation_r", [52], 0.0)
        in validation_errors
    )
    assert (
        NotGreaterOrEqualError(ComponentType.transformer_tap_regulator, "line_drop_compensation_x", [1], 0.0)
        in validation_errors
    )
    assert NotUniqueError(ComponentType.transformer_tap_regulator, "regulated_object", [51, 54]) in validation_errors


def test_fault(input_data):
    validation_errors = validate_input_data(input_data, calculation_type=CalculationType.short_circuit)
    assert validation_errors is not None
    assert InvalidEnumValueError(ComponentType.fault, "fault_type", [50], FaultType) in validation_errors
    assert InvalidEnumValueError(ComponentType.fault, "fault_phase", [50], FaultPhase) in validation_errors
    assert FaultPhaseError(ComponentType.fault, ["fault_type", "fault_phase"], [1, *list(range(32, 51))])
    assert NotGreaterOrEqualError(ComponentType.fault, "r_f", [1], 0) in validation_errors
    assert (
        NotIdenticalError(
            ComponentType.fault,
            "fault_type",
            list(range(32, 51)),
            5 * [0] + 4 * [1] + 4 * [2] + 4 * [3] + [_nan_type(ComponentType.fault, "fault_type"), 4],
        )
        in validation_errors
    )
    assert (
        NotIdenticalError(
            ComponentType.fault,
            "fault_phase",
            list(range(32, 51)),
            list(range(2, 7)) + [0, 4, 5, 6] + 2 * list(range(4)) + [_nan_type(ComponentType.fault, "fault_phase"), 7],
        )
        in validation_errors
    )


def test_validate_input_data_asym_calculation(input_data):
    validation_errors = validate_input_data(input_data, symmetric=False)
    assert validation_errors is not None
    assert NotGreaterThanError(ComponentType.node, "u_rated", [1], 0) in validation_errors
    assert NotUniqueError(ComponentType.node, "id", [2, 2]) in validation_errors
    assert NotBooleanError(ComponentType.line, "from_status", [5]) in validation_errors
    assert NotBooleanError(ComponentType.line, "to_status", [4]) in validation_errors
    assert InvalidIdError(ComponentType.line, "from_node", [4], ComponentType.node) in validation_errors
    assert InvalidIdError(ComponentType.line, "to_node", [5], ComponentType.node) in validation_errors


def test_validate_input_data_invalid_structure():
    with pytest.raises(TypeError, match=r"should be a Numpy structured array"):
        validate_input_data({ComponentType.node: np.array([[1, 10500.0], [2, 10500.0]])}, symmetric=True)


def test_generic_branch_input_data(input_data):
    validation_errors = validate_input_data(input_data, symmetric=True)
    assert validation_errors is not None
    assert NotGreaterThanError(ComponentType.generic_branch, "k", [6], 0) in validation_errors
    assert NotGreaterOrEqualError(ComponentType.generic_branch, "sn", [6], 0) in validation_errors


def test_asym_line_input_data(input_data):
    validation_errors = validate_input_data(input_data, symmetric=True)
    assert validation_errors is not None
    assert NotGreaterThanError(ComponentType.asym_line, "r_aa", [55], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.asym_line, "r_ba", [55], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.asym_line, "r_bb", [55], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.asym_line, "r_ca", [55], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.asym_line, "r_cb", [55], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.asym_line, "r_cc", [55], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.asym_line, "r_na", [55], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.asym_line, "r_nb", [55], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.asym_line, "r_nc", [55], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.asym_line, "r_nn", [55], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.asym_line, "x_aa", [55], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.asym_line, "x_ba", [55], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.asym_line, "x_bb", [55], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.asym_line, "x_ca", [55], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.asym_line, "x_cb", [55], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.asym_line, "x_cc", [55], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.asym_line, "x_na", [55], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.asym_line, "x_nb", [55], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.asym_line, "x_nc", [55], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.asym_line, "x_nn", [55], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.asym_line, "c_aa", [55], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.asym_line, "c_ba", [55], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.asym_line, "c_bb", [55], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.asym_line, "c_ca", [55], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.asym_line, "c_cb", [55], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.asym_line, "c_cc", [55], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.asym_line, "c0", [55], 0) in validation_errors
    assert NotGreaterThanError(ComponentType.asym_line, "c1", [55], 0) in validation_errors
    assert (
        MultiFieldValidationError(
            ComponentType.asym_line, ["r_na", "r_nb", "r_nc", "r_nn", "x_na", "x_nb", "x_nc", "x_nn"], [56, 57]
        )
        in validation_errors
    )
    assert (
        MultiFieldValidationError(
            ComponentType.asym_line, ["c_aa", "c_ba", "c_bb", "c_ca", "c_cb", "c_cc", "c0", "c1"], [56]
        )
        in validation_errors
    )
    assert (
        MultiFieldValidationError(ComponentType.asym_line, ["c_aa", "c_ba", "c_bb", "c_ca", "c_cb", "c_cc"], [58])
        in validation_errors
    )



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

import json

import numpy as np

from power_grid_model import PowerGridModel
from power_grid_model.utils import json_deserialize

input_data = {
    "version": "1.0",
    "type": "input",
    "is_batch": False,
    "attributes": {},
    "data": {
        "sym_load": [{"id": 2, "node": 0, "status": 1, "type": 0, "p_specified": 0, "q_specified": 0}],
        "source": [{"id": 1, "node": 0, "status": 1, "u_ref": 1, "sk": 1e20}],
        "node": [{"id": 0, "u_rated": 10e3}],
    },
}

input_data_json = json.dumps(input_data)


def test_multi_dimensional_batch():
    input_dataset = json_deserialize(input_data_json)
    pgm = PowerGridModel(input_dataset)

    u_rated = 10e3
    u_ref = np.array([0.9, 1.0, 1.1], dtype=np.float64).reshape(-1, 1)
    p_specified = np.array([1e6, 2e6, 3e6, 4e6], dtype=np.float64).reshape(-1, 1)
    q_specified = np.array([0.1e6, 0.2e6, 0.3e6, 0.4e6, 0.5e6], dtype=np.float64).reshape(-1, 1)
    i_source_ref = np.abs(p_specified.reshape(1, -1, 1) + 1j * q_specified.reshape(1, 1, -1)) / (
        u_ref.reshape(-1, 1, 1) * u_rated * np.sqrt(3)
    )
    i_source_ref = i_source_ref.ravel()

    u_ref_batch = {"source": {"u_ref": u_ref}}
    p_specified_batch = {"sym_load": {"p_specified": p_specified}}
    q_specified_batch = {"sym_load": {"q_specified": q_specified}}

    result = pgm.calculate_power_flow(
        update_data=[u_ref_batch, p_specified_batch, q_specified_batch], output_component_types={"source": ["i"]}
    )

    assert np.allclose(result["source"]["i"].ravel(), i_source_ref)



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

from collections.abc import Callable
from copy import copy
from pathlib import Path

import numpy as np
import pytest

from power_grid_model._core.dataset_definitions import ComponentType
from power_grid_model._core.utils import convert_batch_dataset_to_batch_list
from power_grid_model.enum import TapChangingStrategy

from .utils import (
    EXPORT_OUTPUT,
    KNOWN_EXCEPTIONS,
    PowerGridModelWithExt,
    compare_result,
    import_case_data,
    pytest_cases,
    save_json_data,
)

calculation_function_arguments_map: dict[str, tuple[Callable, list[str]]] = {
    "power_flow": (
        PowerGridModelWithExt.calculate_power_flow_with_ext,
        [
            "symmetric",
            "error_tolerance",
            "max_iterations",
            "calculation_method",
            "update_data",
            "threading",
            "output_component_types",
            "continue_on_batch_error",
            "tap_changing_strategy",
            "experimental_features",
        ],
    ),
    "state_estimation": (
        PowerGridModelWithExt.calculate_state_estimation_with_ext,
        [
            "symmetric",
            "error_tolerance",
            "max_iterations",
            "calculation_method",
            "update_data",
            "threading",
            "output_component_types",
            "continue_on_batch_error",
            "experimental_features",
        ],
    ),
    "short_circuit": (
        PowerGridModelWithExt.calculate_short_circuit_with_ext,
        [
            "calculation_method",
            "update_data",
            "threading",
            "output_component_types",
            "continue_on_batch_error",
            "short_circuit_voltage_scaling",
            "experimental_features",
        ],
    ),
}


def supported_kwargs(kwargs, supported: list[str]):
    return {key: value for key, value in kwargs.items() if key in supported}


def get_kwargs(sym: bool, calculation_type: str, calculation_method: str, params: dict, **extra_kwargs) -> dict:
    base_kwargs = {"symmetric": sym, "calculation_method": calculation_method}
    for key, value in params.items():
        if key not in base_kwargs:
            if not isinstance(value, dict):
                base_kwargs[key] = value
            elif calculation_method in value:
                base_kwargs[key] = value[calculation_method]

    base_kwargs.update(params.get("extra_params", {}).get(calculation_method, {}))
    base_kwargs.update(extra_kwargs)

    if calculation_method == "iec60909":
        base_kwargs["short_circuit_voltage_scaling"] = params["short_circuit_voltage_scaling"]

    if calculation_type == "power_flow":
        strategy_name = params.get("tap_changing_strategy", "disabled")
        base_kwargs["tap_changing_strategy"] = TapChangingStrategy[strategy_name].value

    return base_kwargs


def maybe_raises(params: dict):
    raises = params.get("raises")
    if raises:
        raise_type = KNOWN_EXCEPTIONS[raises.get("raises", "AssertionError")]
        if raise_type is not None:
            return pytest.raises(raise_type)

    class PassThrough:
        def __enter__(self):
            pass

        def __exit__(self, exc_type, exc_value, traceback):
            pass

    return PassThrough()


@pytest.mark.parametrize(
    ("case_id", "case_path", "sym", "calculation_type", "calculation_method", "rtol", "atol", "params"),
    pytest_cases(get_batch_cases=False),
)
def test_single_validation(
    case_id: str,
    case_path: Path,
    sym: bool,
    calculation_type: str,
    calculation_method: str,
    rtol: float,
    atol: float,
    params: dict,
):
    with maybe_raises(params):
        # Initialization
        case_data = import_case_data(case_path, calculation_type=calculation_type, sym=sym)
        model = PowerGridModelWithExt(case_data["input"], system_frequency=50.0)

        # Normal calculation
        calculation_function, calculation_args = calculation_function_arguments_map[calculation_type]

        base_kwargs = get_kwargs(
            sym=sym, calculation_type=calculation_type, calculation_method=calculation_method, params=params
        )
        result = calculation_function(model, **supported_kwargs(kwargs=base_kwargs, supported=calculation_args))

        # export data if needed
        if EXPORT_OUTPUT:
            save_json_data(f"{case_id}.json", result)

        # Compare the results
        reference_result = case_data["output"]
        compare_result(result, reference_result, rtol, atol)

        # test get indexer
        for component_name, input_array in case_data["input"].items():
            ids_array = input_array["id"].copy()
            rng = np.random.default_rng(3)
            rng.shuffle(ids_array)
            indexer_array = model.get_indexer(component_name, ids_array)
            # check
            assert np.all(input_array["id"][indexer_array] == ids_array)

        # test calculate with only node and source result
        kwargs = dict(base_kwargs, output_component_types=[ComponentType.node, ComponentType.source])
        result = calculation_function(model, **supported_kwargs(kwargs=kwargs, supported=calculation_args))
        assert set(result.keys()) == {ComponentType.node, ComponentType.source}
        kwargs = dict(base_kwargs, output_component_types={ComponentType.node, ComponentType.source})
        result = calculation_function(model, **supported_kwargs(kwargs=kwargs, supported=calculation_args))
        assert set(result.keys()) == {ComponentType.node, ComponentType.source}


@pytest.mark.parametrize(
    ("case_id", "case_path", "sym", "calculation_type", "calculation_method", "rtol", "atol", "params"),
    pytest_cases(get_batch_cases=True),
)
def test_batch_validation(
    case_id: str,
    case_path: Path,
    sym: bool,
    calculation_type: str,
    calculation_method: str,
    rtol: float,
    atol: float,
    params: dict,
):
    with maybe_raises(params):
        # Initialization
        case_data = import_case_data(case_path, calculation_type=calculation_type, sym=sym)
        model = PowerGridModelWithExt(case_data["input"], system_frequency=50.0)
        update_batch = case_data["update_batch"]
        update_list = convert_batch_dataset_to_batch_list(update_batch)
        reference_output_batch = case_data["output_batch"]
        reference_output_list = convert_batch_dataset_to_batch_list(reference_output_batch)

        base_kwargs = get_kwargs(
            sym=sym, calculation_type=calculation_type, calculation_method=calculation_method, params=params
        )

        calculation_function, calculation_args = calculation_function_arguments_map[calculation_type]

        # export data without comparing first, if needed
        if EXPORT_OUTPUT:
            model_copy = copy(model)
            kwargs = dict(base_kwargs, update_data=update_batch)
            result_batch = calculation_function(
                model_copy, **supported_kwargs(kwargs=kwargs, supported=calculation_args)
            )
            save_json_data(f"{case_id}.json", result_batch)

        # execute batch calculation by applying update method
        for update_data, reference_result in zip(update_list, reference_output_list):
            model_copy = copy(model)
            model_copy.update(update_data=update_data)
            result = calculation_function(
                model_copy, **supported_kwargs(kwargs=base_kwargs, supported=calculation_args)
            )
            compare_result(result, reference_result, rtol, atol)

        # execute in batch one go
        for threading in [-1, 0, 1, 2]:
            kwargs = dict(base_kwargs, update_data=update_batch, threading=threading)
            result_batch = calculation_function(model, **supported_kwargs(kwargs=kwargs, supported=calculation_args))
            result_list = convert_batch_dataset_to_batch_list(result_batch)
            for result, reference_result in zip(result_list, reference_output_list):
                compare_result(result, reference_result, rtol, atol)
